import os
from abc import ABCMeta
from collections import OrderedDict
from uuid import uuid1

import six
from six import with_metaclass

from eventsourcing.exceptions import EventHashError
from eventsourcing.utils.hashing import hash_object
from eventsourcing.utils.times import decimaltimestamp
from eventsourcing.utils.topic import get_topic
from eventsourcing.utils.transcoding import ObjectJSONEncoder

GENESIS_HASH = os.getenv('GENESIS_HASH', '')


class QualnameABCMeta(ABCMeta):
    """
    Supplies __qualname__ to object classes in Python 2.7.

    Needed to get topics from nested classes in Python 2.7.
    """
    __outer_classes = {}

    if not hasattr(object, '__qualname__'):

        def __init__(cls, name, bases, dict):
            super(QualnameABCMeta, cls).__init__(name, bases, dict)
            QualnameABCMeta.__outer_classes[cls] = None
            for class_attr in dict.values():  # iterate class attributes
                for outer_class in QualnameABCMeta.__outer_classes:
                    if class_attr == outer_class:  # is the object an already registered type?
                        QualnameABCMeta.__outer_classes[outer_class] = cls
                        break

        @property
        def __qualname__(cls):
            c = QualnameABCMeta.__outer_classes[cls]
            if c is None:
                return cls.__name__
            else:
                return "%s.%s" % (c.__qualname__, cls.__name__)


def create_timesequenced_event_id():
    return uuid1()


class QualnameABC(with_metaclass(QualnameABCMeta)):
    """
    Base class that introduces __qualname__ for objects in Python 2.7.
    """


class DomainEvent(QualnameABC):
    """
    Base class for domain events.

    Implements methods to make instances read-only, comparable
    for equality, have recognisable representations, and hashable.
    """
    __json_encoder_class__ = ObjectJSONEncoder
    __notifiable__ = True

    def __init__(self, **kwargs):
        """
        Initialises event attribute values directly from constructor kwargs.
        """
        self.__dict__.update(kwargs)

    def __repr__(self):
        """
        Returns string representing the type and attribute values of the event.
        """
        sorted_items = tuple(sorted(self.__dict__.items()))
        args_strings = ("{0}={1!r}".format(*item) for item in sorted_items)
        args_string = ', '.join(args_strings)
        return "{}({})".format(self.__class__.__qualname__, args_string)

    def __mutate__(self, obj):
        """
        Update obj with values from self.

        Can be extended, but subclasses must call super
        method, and return an object.

        :param obj: object to be mutated
        :return: mutated object
        """
        # Call mutate() method.
        self.mutate(obj)

        return obj

    def mutate(self, obj):
        """
        Convenience for use in custom models, to update
        obj with values from self without needing to call
        super method and return obj (two extra lines).

        Can be overridden by subclasses. Any value returned
        by this method will be ignored.

        Please note, subclasses that extend mutate() might
        not have fully completed that method before this method
        is called. To ensure all base classes have completed
        their mutate behaviour before mutating an event in a concrete
        class, extend mutate() instead of overriding this method.

        :param obj: object to be mutated
        """

    def __setattr__(self, key, value):
        """
        Inhibits event attributes from being updated by assignment.
        """
        raise AttributeError("DomainEvent attributes are read-only")

    def __eq__(self, other):
        """
        Tests for equality of two event objects.
        """
        return isinstance(other, DomainEvent) and self.__hash__() == other.__hash__()

    def __ne__(self, other):
        """
        Negates the equality test.
        """
        return not (self == other)

    def __hash__(self):
        """
        Computes a Python integer hash for an event,
        using its event hash string.

        Supports equality and inequality comparisons.
        """
        state = self.__dict__.copy()
        state['__event_topic__'] = get_topic(type(self))
        return hash(self.__hash_object__(state))

    @classmethod
    def __hash_object__(cls, obj):
        return hash_object(cls.__json_encoder_class__, obj)


class EventWithHash(DomainEvent):
    """
    Base class for domain events.

    Implements methods to make instances read-only, comparable
    for equality, have recognisable representations, and hashable.
    """

    def __init__(self, **kwargs):
        """
        Initialises event attribute values directly from constructor kwargs.
        """
        super(EventWithHash, self).__init__(**kwargs)

        # Seal the event with a hash of the other values.
        # __event_topic__ is needed to obtain a different hash
        # for different types of events with otherwise equal
        # attributes.
        self.__dict__['__event_topic__'] = get_topic(type(self))
        self.__dict__['__event_hash__'] = self.__hash_object__(self.__dict__)

    @property
    def __event_hash__(self):
        return self.__dict__.get('__event_hash__')

    def __check_hash__(self):
        state = self.__dict__.copy()
        event_hash = state.pop('__event_hash__')
        if event_hash != self.__hash_object__(state):
            raise EventHashError()

    def __hash__(self):
        """
        Computes a Python integer hash for an event,
        using its event hash string.

        Supports equality and inequality comparisons.
        """
        return hash(self.__event_hash__)

    def __mutate__(self, obj):
        """
        Update obj with values from self.

        Can be extended, but subclasses must call super
        method, and return an object.

        :param obj: object to be mutated
        :return: mutated object
        """

        # Check the hash.
        self.__check_hash__()

        # Call mutate() method.
        return super(EventWithHash, self).__mutate__(obj)


class EventWithOriginatorID(DomainEvent):
    def __init__(self, originator_id, **kwargs):
        kwargs['originator_id'] = originator_id
        super(EventWithOriginatorID, self).__init__(**kwargs)

    @property
    def originator_id(self):
        return self.__dict__['originator_id']


class EventWithTimestamp(DomainEvent):
    """
    For events that have a timestamp value.
    """

    def __init__(self, timestamp=None, **kwargs):
        kwargs['timestamp'] = timestamp or decimaltimestamp()
        super(EventWithTimestamp, self).__init__(**kwargs)

    @property
    def timestamp(self):
        return self.__dict__['timestamp']


class EventWithOriginatorVersion(DomainEvent):
    """
    For events that have an originator version number.
    """

    def __init__(self, originator_version, **kwargs):
        if not isinstance(originator_version, six.integer_types):
            raise TypeError("Version must be an integer: {}".format(originator_version))
        kwargs['originator_version'] = originator_version
        super(EventWithOriginatorVersion, self).__init__(**kwargs)

    @property
    def originator_version(self):
        return self.__dict__['originator_version']


class EventWithTimeuuid(DomainEvent):
    """
    For events that have an UUIDv1 event ID.
    """

    def __init__(self, event_id=None, **kwargs):
        kwargs['event_id'] = event_id or uuid1()
        super(EventWithTimeuuid, self).__init__(**kwargs)

    @property
    def event_id(self):
        return self.__dict__['event_id']


class Created(DomainEvent):
    """
    Can be published when an entity is created.
    """


class AttributeChanged(DomainEvent):
    """
    Can be published when an attribute of an entity is created.
    """

    @property
    def name(self):
        return self.__dict__['name']

    @property
    def value(self):
        return self.__dict__['value']


class Discarded(DomainEvent):
    """
    Published when something is discarded.
    """


class Logged(DomainEvent):
    """
    Published when something is logged.
    """


_event_handlers = OrderedDict()


def subscribe(handler, predicate=None):
    if predicate not in _event_handlers:
        _event_handlers[predicate] = []
    _event_handlers[predicate].append(handler)


def unsubscribe(handler, predicate=None):
    if predicate in _event_handlers:
        handlers = _event_handlers[predicate]
        if handler in handlers:
            handlers.remove(handler)
            if not handlers:
                _event_handlers.pop(predicate)


def publish(event):
    for predicate, handlers in _event_handlers.copy().items():
        for handler in handlers:
            if predicate is None or predicate(event):
                handler(event)


class EventHandlersNotEmptyError(Exception):
    pass


def assert_event_handlers_empty():
    if len(_event_handlers):
        msg = "Event handlers are still subscribed: %s" % _event_handlers
        raise EventHandlersNotEmptyError(msg)


def clear_event_handlers():
    _event_handlers.clear()



from collections import OrderedDict, defaultdict

from eventsourcing.application.pipeline import Pipeable
from eventsourcing.application.simple import SimpleApplication
from eventsourcing.application.snapshotting import ApplicationWithSnapshotting
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.entity import DomainEntity
from eventsourcing.domain.model.events import publish, subscribe, unsubscribe
from eventsourcing.exceptions import CausalDependencyFailed, OperationalError, PromptFailed, RecordConflictError
from eventsourcing.infrastructure.base import RelationalRecordManager
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.interface.notificationlog import NotificationLogReader
from eventsourcing.utils.transcoding import json_dumps, json_loads


class ProcessApplication(Pipeable, SimpleApplication):
    tracking_record_manager_class = None
    always_track_notifications = False

    def __init__(self, name=None, policy=None, setup_table=False, tracking_record_manager_class=None,
                 always_track_notifications=False, **kwargs):
        self.always_track_notifications = always_track_notifications or self.always_track_notifications
        self.tracking_record_manager_class = tracking_record_manager_class or self.tracking_record_manager_class
        self.policy_func = policy
        self.readers = OrderedDict()
        self.is_reader_position_ok = defaultdict(bool)
        super(ProcessApplication, self).__init__(name=name, setup_table=setup_table, **kwargs)
        # self._cached_entities = {}

        ## Prompts policy.
        #
        # 1. Publish prompts whenever domain events are published (important: after persisted).
        # 2. Run this process whenever upstream prompted followers to pull for new notification.
        subscribe(predicate=self.persistence_policy.is_event, handler=self.publish_prompt_from_event)
        subscribe(predicate=self.is_upstream_prompt, handler=self.run)
        # Todo: Maybe make a prompts policy object?

    def construct_infrastructure_factory(self, *args, **kwargs):
        return super(ProcessApplication, self).construct_infrastructure_factory(
            tracking_record_manager_class=self.tracking_record_manager_class, *args, **kwargs
        )

    def setup_infrastructure(self, *args, **kwargs):
        super(ProcessApplication, self).setup_infrastructure(*args, **kwargs)
        self.tracking_record_manager = self.infrastructure_factory.construct_tracking_record_manager()

    def follow(self, upstream_application_name, notification_log):
        # Create a reader.
        reader = NotificationLogReader(notification_log)
        self.readers[upstream_application_name] = reader

    def setup_table(self):
        super(ProcessApplication, self).setup_table()
        if self.datastore is not None:
            self.datastore.setup_table(
                self.tracking_record_manager.record_class
            )

    @retry((OperationalError, RecordConflictError), max_attempts=100, wait=0.01)
    def run(self, prompt=None, advance_by=None):

        if prompt is None:
            readers_items = self.readers.items()
        else:
            assert isinstance(prompt, Prompt)
            reader = self.readers[prompt.process_name]
            readers_items = [(prompt.process_name, reader)]

        notification_count = 0
        for upstream_application_name, reader in readers_items:

            if not self.is_reader_position_ok[upstream_application_name]:
                self.set_reader_position_from_tracking_records(reader, upstream_application_name)
                self.is_reader_position_ok[upstream_application_name] = True

            for notification in reader.read(advance_by=advance_by):
                # Todo: Put this on a queue and then get the next one.
                notification_count += 1

                # Todo: Get this from a queue and do it in a different thread?
                # Domain event from notification.
                event = self.event_store.sequenced_item_mapper.from_topic_and_data(
                    topic=notification['event_type'],
                    data=notification['state']
                )

                # Wait for causal dependencies to be satisfied.
                upstream_causal_dependencies = notification.get('causal_dependencies')
                if upstream_causal_dependencies is not None:
                    upstream_causal_dependencies = json_loads(upstream_causal_dependencies)
                if upstream_causal_dependencies is None:
                    upstream_causal_dependencies = []
                for causal_dependency in upstream_causal_dependencies:
                    pipeline_id = causal_dependency['pipeline_id']
                    notification_id = causal_dependency['notification_id']

                    if not self.tracking_record_manager.has_tracking_record(
                        application_name=self.name,
                        upstream_application_name=upstream_application_name,
                        pipeline_id=pipeline_id,
                        notification_id=notification_id
                    ):
                        self.is_reader_position_ok[upstream_application_name] = False

                        raise CausalDependencyFailed({
                            'application_name': self.name,
                            'upstream_application_name': upstream_application_name,
                            'pipeline_id': pipeline_id,
                            'notification_id': notification_id
                        })

                # Call policy with the event.
                all_aggregates, causal_dependencies = self.call_policy(event)

                # Record new events.
                try:
                    new_events = self.record_new_events(
                        all_aggregates, notification, upstream_application_name, causal_dependencies
                    )
                except Exception as e:
                    self.is_reader_position_ok[upstream_application_name] = False
                    # self._cached_entities = {}
                    raise e
                else:
                    # Publish a prompt if there are new notifications.
                    # Todo: Optionally send events as prompts, saves pulling event if it arrives in order.
                    if len(new_events):
                        self.publish_prompt()

        return notification_count

    def set_reader_position_from_tracking_records(self, reader, upstream_application_name):
        max_record_id = self.tracking_record_manager.get_max_record_id(
            application_name=self.name,
            upstream_application_name=upstream_application_name,
            pipeline_id=self.pipeline_id,
        )
        reader.seek(max_record_id or 0)

    def call_policy(self, event):
        repository = RepositoryWrapper(self.repository)
        new_aggregates = self.policy(repository, event)
        repo_aggregates = list(repository.retrieved_aggregates.values())
        all_aggregates = repo_aggregates[:]
        if new_aggregates is not None:
            if not isinstance(new_aggregates, (list, tuple)):
                new_aggregates = [new_aggregates]
            all_aggregates += new_aggregates

        highest = defaultdict(int)
        for entity_id, entity_version in repository.causal_dependencies:
            pipeline_id, notification_id = self.event_store.record_manager.get_pipeline_and_notification_id(
                entity_id, entity_version
            )
            if pipeline_id is not None and pipeline_id != self.pipeline_id:
                highest[pipeline_id] = max(notification_id, highest[pipeline_id])

        causal_dependencies = []
        for pipeline_id, notification_id in highest.items():
            causal_dependencies.append({
                'pipeline_id': pipeline_id,
                'notification_id': notification_id
            })
        # Todo: Optionally reference causal dependencies in current pipeline.
        # Todo: Support processing notification from a single pipeline in parallel, according to dependencies.
        return all_aggregates, causal_dependencies

    def policy(self, repository, event):
        return self.policy_func(self, repository, event) if self.policy_func is not None else None

    def record_new_events(self, aggregates, notification=None, upstream_application_name=None,
                          causal_dependencies=None):
        # Construct tracking record.
        tracking_kwargs = self.construct_tracking_kwargs(notification, upstream_application_name)

        # Collect pending events.
        new_events, num_changed_aggregates = self.collect_pending_events(aggregates)

        # Writing tracking record when there are new events.
        if len(new_events) or self.always_track_notifications:

            # Sort pending events across all aggregates.
            if num_changed_aggregates > 1:
                self.sort_pending_events(new_events)

            # Construct event records.
            event_records = self.construct_event_records(new_events, causal_dependencies)

            # Write event records with tracking record.
            record_manager = self.event_store.record_manager
            assert isinstance(record_manager, RelationalRecordManager)
            record_manager.write_records(records=event_records, tracking_kwargs=tracking_kwargs)
        # else:
            # Todo: Maybe write one tracking record at the end of a run, if necessary, or
            # only during a period of time when nothing happens?


        return new_events

    def sort_pending_events(self, pending_events):
        # Sort the events by timestamp.
        #  - this approximation is a supposed to correlate with the correct
        #    causal ordering of all new events across all aggregates. It
        #    should work if all events are timestamped, all their timestamps
        #    are from the same clock, and none have the same value. If this
        #    doesn't work properly, it is possible when several aggregates
        #    publish that depend on each other that concatenating pending events
        #    taken from each in turn will be incorrect and could potentially
        #    cause processing errors in a downstream process application that
        #    somehow depends on the correct ordering of events.
        try:
            pending_events.sort(key=lambda x: x.timestamp)
        except AttributeError:
            pass

    def is_upstream_prompt(self, prompt):
        return isinstance(prompt, Prompt) and prompt.process_name in self.readers.keys()

    def publish_prompt_from_event(self, _):
        self.publish_prompt()

    def publish_prompt(self):
        prompt = Prompt(self.name, self.pipeline_id)
        try:
            publish(prompt)
        except PromptFailed:
            raise
        except Exception as e:
            raise PromptFailed("{}: {}".format(type(e), str(e)))

    def construct_event_records(self, pending_events, causal_dependencies=None):
        # Convert to event records.
        sequenced_items = self.event_store.to_sequenced_item(pending_events)
        event_records = self.event_store.record_manager.to_records(sequenced_items)

        # Set notification log IDs, and causal dependencies.
        if len(event_records):
            current_max = self.event_store.record_manager.get_max_record_id() or 0
            for domain_event, event_record in zip(pending_events, event_records):
                if type(domain_event).__notifiable__:
                    current_max += 1
                    event_record.id = current_max
                else:
                    event_record.id = ''

            # Only need first event to carry the dependencies.
            if hasattr(self.event_store.record_manager.record_class, 'causal_dependencies'):
                causal_dependencies = json_dumps(causal_dependencies)
                event_records[0].causal_dependencies = causal_dependencies

        return event_records

    def collect_pending_events(self, aggregates):
        if isinstance(aggregates, DomainEntity):
            aggregates = [aggregates]
        assert isinstance(aggregates, (list, tuple))
        pending_events = []
        num_changed_aggregates = 0
        for aggregate in aggregates:
            batch = aggregate.__batch_pending_events__()
            if len(batch):
                num_changed_aggregates += 1
            pending_events += batch
        return pending_events, num_changed_aggregates

    def construct_tracking_kwargs(self, notification, upstream_application_name):
        if notification is None:
            return {}
        tracking_kwargs = {
            'application_name': self.name,
            'upstream_application_name': upstream_application_name,
            'pipeline_id': self.pipeline_id,
            'notification_id': notification['id'],
        }
        return tracking_kwargs

    def close(self):
        unsubscribe(predicate=self.is_upstream_prompt, handler=self.run)
        unsubscribe(predicate=self.persistence_policy.is_event, handler=self.publish_prompt_from_event)
        super(ProcessApplication, self).close()


class RepositoryWrapper(object):
    def __init__(self, repository):
        self.retrieved_aggregates = {}
        assert isinstance(repository, EventSourcedRepository)
        self.repository = repository
        self.causal_dependencies = []

    def __getitem__(self, entity_id):
        try:
            return self.retrieved_aggregates[entity_id]
        except KeyError:
            entity = self.repository.__getitem__(entity_id)
            self.retrieved_aggregates[entity_id] = entity
            self.causal_dependencies.append((entity.id, entity.__version__))
            return entity

    def __contains__(self, entity_id):
        return self.repository.__contains__(entity_id)


class Prompt(object):
    def __init__(self, process_name, pipeline_id):
        self.process_name = process_name
        self.pipeline_id = pipeline_id


class ProcessApplicationWithSnapshotting(ApplicationWithSnapshotting, ProcessApplication):
    def record_new_events(self, *args, **kwargs):
        new_events = super(ProcessApplicationWithSnapshotting, self).record_new_events(*args, **kwargs)
        for event in new_events:
            if self.snapshotting_policy.condition(event):
                self.snapshotting_policy.take_snapshot(event)
        return new_events



from uuid import uuid4

from time import sleep

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.example.application import ExampleApplication
from eventsourcing.example.domainmodel import Example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.exceptions import ProgrammingError
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.tests.sequenced_item_tests.base import WithRecordManagers


class WithExampleApplication(WithRecordManagers):
    def construct_application(self):
        cipher = self.construct_cipher()
        app = ExampleApplication(
            entity_record_manager=self.entity_record_manager,
            log_record_manager=self.log_record_manager,
            snapshot_record_manager=self.snapshot_record_manager,
            cipher=cipher,
        )
        return app

    def construct_cipher(self):
        return None


class ExampleApplicationTestCase(WithExampleApplication):
    drop_tables = True

    def test(self):
        """
        Checks the example application works in the way an example application should.
        """

        with self.construct_application() as app:

            # Check there's an event store for entity events.
            self.assertIsInstance(app.entity_event_store, EventStore)

            # Check there's an event store for log events.
            self.assertIsInstance(app.log_event_store, EventStore)

            # Check there's a persistence policy.
            self.assertIsInstance(app.entity_persistence_policy, PersistencePolicy)

            # Check there's an example repository.
            self.assertIsInstance(app.example_repository, ExampleRepository)

            # Take snapshot of non-existing entity.
            snapshot0 = app.example_repository.take_snapshot(uuid4())
            self.assertIsNone(snapshot0)

            # Register a new example.
            example1 = app.create_new_example(a=10, b=20)
            self.assertIsInstance(example1, Example)

            # Check the example is available in the repo.
            entity1 = app.example_repository[example1.id]
            self.assertEqual(10, entity1.a)
            self.assertEqual(20, entity1.b)
            self.assertEqual(example1, entity1)

            # Change attribute values.
            entity1.a = 50

            # Check the attribute has the new value.
            self.assertEqual(50, entity1.a)

            # Check the new value is available in the repo.
            entity1 = app.example_repository[example1.id]
            self.assertEqual(50, entity1.a)

            # Take a snapshot of the entity.
            snapshot1 = app.example_repository.take_snapshot(entity1.id)
            self.assertEqual(snapshot1.originator_id, entity1.id)
            self.assertEqual(snapshot1.originator_version, entity1.__version__)

            # Check the snapshot exists.
            snapshot2 = app.snapshot_strategy.get_snapshot(entity1.id)
            self.assertIsInstance(snapshot2, Snapshot)
            self.assertEqual(snapshot1, snapshot2)

            # Take another snapshot of the entity (should be the same event).
            sleep(0.0001)
            snapshot2 = app.example_repository.take_snapshot(entity1.id)
            self.assertEqual(snapshot1, snapshot2)

            # Change attribute values.
            entity1.a = 100

            # Check the attribute has the new value.
            self.assertEqual(100, entity1.a)

            # Check the new value is available in the repo.
            entity1 = app.example_repository[example1.id]
            self.assertEqual(100, entity1.a)

            # Check the old value is available in the repo.
            entity1_v1 = app.example_repository.get_entity(entity1.id, at=0)
            self.assertEqual(entity1_v1.a, 10)
            entity1_v2 = app.example_repository.get_entity(entity1.id, at=1)
            self.assertEqual(entity1_v2.a, 50)
            entity1_v3 = app.example_repository.get_entity(entity1.id, at=2)
            self.assertEqual(entity1_v3.a, 100)

            # Take another snapshot of the entity.
            snapshot4 = app.example_repository.take_snapshot(entity1.id)

            # Check the new snapshot is not equal to the first.
            self.assertNotEqual(snapshot1, snapshot4)

            # Check the new value still available in the repo.
            self.assertEqual(100, app.example_repository[example1.id].a)

            # Remove all the stored items and check the new value is still available (must be in snapshot).
            record_strategy = self.entity_record_manager
            records = record_strategy.get_records(example1.id)
            self.assertEqual(len(records), 3)
            for record in records:
                record_strategy.delete_record(record)
            self.assertFalse(record_strategy.get_records(example1.id))
            self.assertEqual(100, app.example_repository[example1.id].a)

            # Check only some of the old values are available in the repo.
            entity1_v1 = app.example_repository.get_entity(entity1.id, at=0)
            self.assertEqual(entity1_v1, None)
            entity1_v3 = app.example_repository.get_entity(entity1.id, at=1)
            self.assertEqual(entity1_v3.a, 50)
            entity1_v3 = app.example_repository.get_entity(entity1.id, at=2)
            self.assertEqual(entity1_v3.a, 100)

            # Test 'except' clause in delete_record() method.
            # - register a new example.
            example1 = app.create_new_example(a=10, b=20)
            self.assertIsInstance(example1, Example)

            # - get the records to delete
            records = record_strategy.get_records(example1.id)
            self.assertEqual(1, len(records))

            # - drop the table...
            if self.datastore:
                self.datastore.drop_table(record_strategy.record_class)

            # - check exception is raised when records can't be deleted, so that
            #   test case runs through 'except' block, so rollback() is called
            for record in records:
                try:
                    record_strategy.delete_record(record)
                except ProgrammingError:
                    pass



import importlib

from eventsourcing.exceptions import TopicResolutionError


def get_topic(domain_class):
    """Returns a string describing a class.

    Args:
        domain_class: A class.

    Returns:
        A string describing the class.
    """
    return domain_class.__module__ + '#' + getattr(domain_class, '__qualname__', domain_class.__name__)


def resolve_topic(topic):
    """Return class described by given topic.

    Args:
        topic: A string describing a class.

    Returns:
        A class.

    Raises:
        TopicResolutionError: If there is no such class.
    """
    try:
        module_name, _, class_name = topic.partition('#')
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    try:
        cls = resolve_attr(module, class_name)
    except AttributeError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    return cls


def resolve_attr(obj, path):
    """A recursive version of getattr for navigating dotted paths.

    Args:
        obj: An object for which we want to retrieve a nested attribute.
        path: A dot separated string containing zero or more attribute names.

    Returns:
        The attribute referred to by obj.a1.a2.a3...

    Raises:
        AttributeError: If there is no such attribute.
    """
    if not path:
        return obj
    head, _, tail = path.partition('.')
    head_obj = getattr(obj, head)
    return resolve_attr(head_obj, tail)



from abc import ABCMeta, abstractmethod

import six

from eventsourcing.exceptions import OperationalError, RecordConflictError
from eventsourcing.infrastructure.sequenceditem import SequencedItem, SequencedItemFieldNames


class AbstractSequencedItemRecordManager(six.with_metaclass(ABCMeta)):
    def __init__(self, record_class, sequenced_item_class=SequencedItem, contiguous_record_ids=False,
                 application_name=None, pipeline_id=-1):
        self.record_class = record_class
        self.sequenced_item_class = sequenced_item_class
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)
        self.contiguous_record_ids = contiguous_record_ids and hasattr(self.record_class, 'id')
        if hasattr(self.record_class, 'application_name'):
            assert application_name, "'application_name' not set when required"
            assert contiguous_record_ids, "'contiguous_record_ids' not set when required"
        self.application_name = application_name
        if hasattr(self.record_class, 'pipeline_id'):
            assert hasattr(self.record_class, 'application_name'), "'application_name' column not defined"
        self.pipeline_id = pipeline_id

    @abstractmethod
    def append(self, sequenced_item_or_items):
        """
        Writes sequenced item into the datastore.
        """

    @abstractmethod
    def get_item(self, sequence_id, position):
        """
        Gets sequenced item from the datastore.
        """

    def list_items(self, *args, **kwargs):
        return list(self.get_items(*args, **kwargs))

    def get_items(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                  query_ascending=True, results_ascending=True):
        """
        Returns sequenced items.
        """
        records = self.get_records(
            sequence_id=sequence_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            query_ascending=query_ascending,
            results_ascending=results_ascending,

        )
        for item in six.moves.map(self.from_record, records):
            yield item

    @abstractmethod
    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):
        """
        Returns records for a sequence.
        """

    def to_record(self, sequenced_item):
        """
        Constructs and returns an ORM object, from given sequenced item object.
        """
        kwargs = self.get_field_kwargs(sequenced_item)
        if hasattr(self.record_class, 'application_name'):
            kwargs['application_name'] = self.application_name
        return self.record_class(**kwargs)

    def from_record(self, record):
        """
        Constructs and returns a sequenced item object, from given ORM object.
        """
        kwargs = self.get_field_kwargs(record)
        return self.sequenced_item_class(**kwargs)

    @abstractmethod
    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        """
        Returns records sequenced by notification ID, from application, for pipeline, in given range.
        """

    @abstractmethod
    def all_sequence_ids(self):
        """
        Returns all sequence IDs.
        """

    def list_sequence_ids(self):
        return list(self.all_sequence_ids())

    @abstractmethod
    def delete_record(self, record):
        """
        Removes permanently given record from the table.
        """

    def get_field_kwargs(self, item):
        return {name: getattr(item, name) for name in self.field_names}

    def raise_sequenced_item_conflict(self):
        msg = "Position already taken in sequence"
        raise RecordConflictError(msg)

    def raise_index_error(self, position):
        raise IndexError("Sequence index out of range: {}".format(position))

    def raise_record_integrity_error(self, e):
        raise RecordConflictError(e)

    def raise_operational_error(self, e):
        raise OperationalError(e)


class ACIDRecordManager(AbstractSequencedItemRecordManager):
    """
    ACID record managers can write tracking records and event records
    in an atomic transaction, needed for atomic processing in process
    applications.
    """
    tracking_record_class = None

    tracking_record_field_names = [
        'application_name',
        'upstream_application_name',
        'pipeline_id',
        'notification_id',
        # 'originator_id',
        # 'originator_version',
    ]

    @abstractmethod
    def get_max_record_id(self):
        """Return maximum notification ID in pipeline."""


class RelationalRecordManager(ACIDRecordManager):
    def __init__(self, *args, **kwargs):
        super(RelationalRecordManager, self).__init__(*args, **kwargs)
        self._insert_select_max = None
        self._insert_values = None
        self._insert_tracking_record = None

    def append(self, sequenced_item_or_items):
        # Convert sequenced item(s) to database record(s).
        records = self.to_records(sequenced_item_or_items)

        # Write records.
        self.write_records(records)

    def to_records(self, sequenced_item_or_items):
        if isinstance(sequenced_item_or_items, list):
            records = [self.to_record(i) for i in sequenced_item_or_items]
        else:
            records = [self.to_record(sequenced_item_or_items)]
        return records

    def write_records(self, records, tracking_kwargs=None):
        """
        Creates records in the database.
        :param tracking_kwargs:
        """

    @property
    def insert_select_max(self):
        """
        SQL statement that inserts records with contiguous IDs,
        by selecting max ID from indexed table records.
        """
        if self._insert_select_max is None:
            if hasattr(self.record_class, 'application_name'):
                # Todo: Maybe make it support application_name with pipeline_id?
                assert hasattr(self.record_class, 'pipeline_id')
                tmpl = self._insert_select_max_where_application_name_tmpl
            else:
                tmpl = self._insert_select_max_tmpl
            self._insert_select_max = self._prepare_insert(
                tmpl=tmpl,
                record_class=self.record_class,
                field_names=list(self.field_names),
            )
        return self._insert_select_max

    @abstractmethod
    def _prepare_insert(self, tmpl, record_class, field_names, placeholder_for_id=False):
        """
        Compile SQL statement with placeholders for bind parameters.
        """

    _insert_select_max_tmpl = (
        "INSERT INTO {tablename} (id, {columns}) "
        "SELECT COALESCE(MAX({tablename}.id), 0) + 1, {placeholders} "
        "FROM {tablename};"
    )

    _insert_select_max_where_application_name_tmpl = (
        "INSERT INTO {tablename} (id, {columns}) "
        "SELECT COALESCE(MAX({tablename}.id), 0) + 1, {placeholders} "
        "FROM {tablename} WHERE application_name=:application_name AND pipeline_id=:pipeline_id;"
    )

    @property
    def insert_values(self):
        """
        SQL statement that inserts records without ID.
        """
        if self._insert_values is None:
            self._insert_values = self._prepare_insert(
                tmpl=self._insert_values_tmpl,
                placeholder_for_id=True,
                record_class=self.record_class,
                field_names=self.field_names,
            )
        return self._insert_values

    @property
    def insert_tracking_record(self):
        """
        SQL statement that inserts tracking records.
        """
        if self._insert_tracking_record is None:
            self._insert_tracking_record = self._prepare_insert(
                tmpl=self._insert_values_tmpl,
                placeholder_for_id=True,
                record_class=self.tracking_record_class,
                field_names=self.tracking_record_field_names,
            )
        return self._insert_tracking_record

    _insert_values_tmpl = (
        "INSERT INTO {tablename} ({columns}) "
        "VALUES ({placeholders});"
    )

    @abstractmethod
    def get_record_table_name(self, record_class):
        """
        Returns table name - used in raw queries.

        :rtype: str
        """

    def clone(self, application_name, pipeline_id, **kwargs):
        return type(self)(
            record_class=self.record_class,
            contiguous_record_ids=self.contiguous_record_ids,
            sequenced_item_class=self.sequenced_item_class,
            application_name=application_name,
            pipeline_id=pipeline_id,
            **kwargs
        )


class AbstractTrackingRecordManager(six.with_metaclass(ABCMeta)):

    @property
    @abstractmethod
    def record_class(self):
        """Returns tracking record class."""

    @abstractmethod
    def get_max_record_id(self, application_name, upstream_application_name, pipeline_id):
        """Returns maximum record ID for given application name."""



import datetime
from decimal import Decimal
import time
from uuid import UUID

import six

if hasattr(datetime, 'timezone'):
    utc_timezone = datetime.timezone.utc
else:
    from datetime import tzinfo, timedelta


    class UTC(tzinfo):
        def utcoffset(self, date_time):
            return timedelta(0)

        def dst(self, date_time):
            return timedelta(0)


    utc_timezone = UTC()


def decimaltimestamp_from_uuid(uuid_arg):
    """
    Return a floating point unix timestamp.

    :param uuid_arg:
    :return: Unix timestamp in seconds, with microsecond precision.
    :rtype: float
    """
    return decimaltimestamp(timestamp_long_from_uuid(uuid_arg) / 1e7)


def timestamp_long_from_uuid(uuid_arg):
    """
    Returns an integer value representing a unix timestamp in tenths of microseconds.

    :param uuid_arg:
    :return: Unix timestamp integer in tenths of microseconds.
    :rtype: int
    """
    if isinstance(uuid_arg, six.string_types):
        uuid_arg = UUID(uuid_arg)
    assert isinstance(uuid_arg, UUID), uuid_arg
    uuid_time = uuid_arg.time
    return uuid_time - 0x01B21DD213814000


def decimaltimestamp(t=None):
    """
    A UNIX timestamp as a Decimal object (exact number type).

    Returns current time when called without args, otherwise
    converts given floating point number ``t`` to a Decimal
    with 9 decimal places.

    :param t: Floating point UNIX timestamp ("seconds since epoch").
    :return: A Decimal with 6 decimal places, representing the
            given floating point or the value returned by time.time().
    """
    t = time.time() if t is None else t
    return Decimal('{:.6f}'.format(t))
    # return Decimal('{:.9f}'.format(t))


def datetime_from_timestamp(t):
    """
    Returns naive UTC datetime from decimal UNIX
    timestamps such as time.time().

    :param t: timestamp, either Decimal or float
    :return: datetime.datetime object
    """
    return datetime.datetime.utcfromtimestamp(float(t))



from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.entity import TimeuuidedEntity
from eventsourcing.domain.model.events import EventWithTimeuuid, DomainEvent
from eventsourcing.infrastructure.cassandra.datastore import CassandraDatastore, CassandraSettings
from eventsourcing.infrastructure.cassandra.records import TimeuuidSequencedRecord
from eventsourcing.infrastructure.cassandra.manager import CassandraRecordManager
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase
from eventsourcing.tests.datastore_tests.test_cassandra import DEFAULT_KEYSPACE_FOR_TESTING
from eventsourcing.utils.times import decimaltimestamp_from_uuid


# This test has events with TimeUUID value as the 'event ID'. How easy is it to customize
# the infrastructure to support that? We just need to make a model that uses these events,
# define a suitable database table, and configure the other components. It's easy.

# Firstly, define and entity that uses events with TimeUUIDs.


class ExampleEntity(TimeuuidedEntity):
    def __init__(self, **kwargs):
        super(ExampleEntity, self).__init__(**kwargs)
        self._is_finished = False

    class Started(TimeuuidedEntity.Created, EventWithTimeuuid):
        pass

    class Finished(EventWithTimeuuid, TimeuuidedEntity.Discarded):
        pass

    def finish(self):
        self.__trigger_event__(self.Finished)

    @classmethod
    def start(cls):
        return cls.__create__(event_class=ExampleEntity.Started)


# Define a suitable record class.
#  - class TimeuuidSequencedItem has been moved into the library

# Define an application that uses the entity class and the table
class ExampleApplicationWithTimeuuidSequencedItems(object):
    def __init__(self):
        self.event_store = EventStore(
            record_manager=CassandraRecordManager(
                record_class=TimeuuidSequencedRecord,
                sequenced_item_class=SequencedItem,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name='originator_id',
                position_attr_name='event_id',
            )
        )
        self.repository = EventSourcedRepository(
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store,
            event_type=DomainEvent,
        )

    def start_entity(self):
        return ExampleEntity.start()

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestDomainEventsWithTimeUUIDs(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestDomainEventsWithTimeUUIDs, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestDomainEventsWithTimeUUIDs, self).setUp()

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
            tables=(TimeuuidSequencedRecord,),
        )

    def test(self):
        with ExampleApplicationWithTimeuuidSequencedItems() as app:
            # Create entity.
            entity1 = app.start_entity()
            self.assertIsInstance(entity1.___initial_event_id__, UUID)
            expected_timestamp = decimaltimestamp_from_uuid(entity1.___initial_event_id__)
            self.assertEqual(entity1.__created_on__, expected_timestamp)
            self.assertTrue(entity1.__last_modified__, expected_timestamp)

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)

            retrieved_obj.finish()
            assert retrieved_obj.id not in app.repository



from abc import abstractmethod

from eventsourcing.infrastructure.datastore import DatastoreConnectionError, DatastoreTableError
from eventsourcing.tests.base import AbstractTestCase


class AbstractDatastoreTestCase(AbstractTestCase):
    infrastructure_factory_class = None
    contiguous_record_ids = False

    def __init__(self, *args, **kwargs):
        super(AbstractDatastoreTestCase, self).__init__(*args, **kwargs)
        self._datastore = None
        self._factory = None

    def tearDown(self):
        self._datastore = None
        super(AbstractDatastoreTestCase, self).tearDown()

    @property
    def datastore(self):
        """
        :rtype: eventsourcing.infrastructure.datastore.datastore.Datastore
        """
        if self._datastore is None:
            self._datastore = self.construct_datastore()
        return self._datastore

    @abstractmethod
    def construct_datastore(self):
        """
        :rtype: eventsourcing.infrastructure.datastore.datastore.Datastore
        """

    @property
    def factory(self):
        if not self._factory:
            kwargs = {}
            if hasattr(self.datastore, 'session'):
                kwargs['session'] = self.datastore.session
            if self.contiguous_record_ids:
                kwargs['contiguous_record_ids'] = True
            self._factory = self.infrastructure_factory_class(**kwargs)
        return self._factory


class DatastoreTestCase(AbstractDatastoreTestCase):
    def test(self):

        # # Check the stored event class doesn't function before the connection is setup.
        # with self.assertRaises(DatastoreConnectionError):
        #     self.list_records()
        # with self.assertRaises(DatastoreConnectionError):
        #     self.create_record()

        # Setup the connection.
        self.datastore.setup_connection()

        # Check it doesn't matter if setup_connection() is called twice.
        self.datastore.setup_connection()

        # # Check the stored event class doesn't function before the tables are setup.
        # with self.assertRaises(DatastoreTableError):
        #     self.list_records()
        # with self.assertRaises(DatastoreTableError):
        #     self.create_record()

        # Setup the tables.
        self.datastore.setup_tables()

        # Check it doesn't matter if setup_tables() is called twice.
        self.datastore.setup_tables()

        # Check the stored event class does function after the tables have been setup.
        self.assertEqual(len(self.list_records()), 0)
        self.create_record()
        self.assertEqual(len(self.list_records()), 1)

        # Drop the tables.
        self.datastore.drop_tables()

        # Check the stored event class doesn't function after the tables have been dropped.
        with self.assertRaises(DatastoreTableError):
            self.list_records()
        with self.assertRaises(DatastoreTableError):
            self.create_record()

        # Drop the connection.
        self.datastore.close_connection()

    @abstractmethod
    def list_records(self):
        return []

    @abstractmethod
    def create_record(self):
        return None

    def tearDown(self):
        # Try to remove any tables.
        self.datastore.setup_connection()
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(DatastoreTestCase, self).tearDown()



# from unittest.case import TestCase
#
# from eventsourcing.domain.services.retries import retry_after_concurrency_error
# from eventsourcing.exceptions import ConcurrencyError
#
#
# class TestRetries(TestCase):
#
#     def setUp(self):
#         self.call_count = 0
#
#     def test_retry_after_concurrency_error_success(self):
#         retry_after_concurrency_error(
#             operation_to_retry=self.raise_concurrency_error,
#         )
#
#     def test_retry_after_concurrency_error_fail(self):
#         with self.assertRaises(ConcurrencyError):
#             retry_after_concurrency_error(
#                 operation_to_retry=self.raise_concurrency_error,
#                 max_attempts=1,
#             )
#
#
#     def raise_concurrency_error(self):
#         if self.call_count < 2:
#             self.call_count += 1
#             raise ConcurrencyError("Just a test")



# coding=utf-8
from __future__ import unicode_literals

import datetime
from uuid import uuid4

import six

from eventsourcing.domain.model.entity import AbstractEntityRepository, AttributeChanged, Created, Discarded, \
    TimestampedVersionedEntity
from eventsourcing.domain.model.decorators import attribute
from eventsourcing.domain.model.events import publish
from eventsourcing.example.application import ExampleApplication
from eventsourcing.exceptions import RepositoryKeyError


class SuffixTree(TimestampedVersionedEntity):
    """A suffix tree for string matching. Uses Ukkonen's algorithm
    for construction.
    """

    class Created(Created):
        pass

    class AttributeChanged(AttributeChanged):
        pass

    class Discarded(Discarded):
        pass

    def __init__(self, root_node_id, case_insensitive=False, **kwargs):
        """
        string
            the string for which to construct a suffix tree
        """
        super(SuffixTree, self).__init__(**kwargs)
        self._root_node_id = root_node_id
        self._case_insensitive = case_insensitive
        self._nodes = {}
        self._edges = {}
        self._N = None
        self._string = None

    @attribute
    def string(self):
        return self._string

    @property
    def N(self):
        return self._N

    @property
    def nodes(self):
        return self._nodes

    @property
    def root_node_id(self):
        return self._root_node_id

    @property
    def edges(self):
        return self._edges

    @property
    def active(self):
        return self._active

    @property
    def case_insensitive(self):
        return self._case_insensitive

    def __repr__(self):
        """
        Lists edges in the suffix tree
        """
        return "SuffixTree({})".format(self.id)

    def add_string(self, string):
        assert self.string is None
        self._active = Suffix(self._root_node_id, 0, -1)
        self._N = len(string) - 1
        if self._case_insensitive:
            string = string.lower()
        self.string = string
        for i in range(len(string)):
            self._add_prefix(i)

    def _add_prefix(self, last_char_index):
        """The core construction method.
        """
        last_parent_node_id = None
        while True:
            parent_node_id = self.active.source_node_id
            if self.active.explicit():
                edge_id = make_edge_id(self.active.source_node_id, self.string[last_char_index])
                if edge_id in self.edges:
                    # prefix is already in tree
                    break
            else:
                edge_id = make_edge_id(self.active.source_node_id, self.string[self.active.first_char_index])
                e = self.edges[edge_id]
                if self.string[e.first_char_index + self.active.length + 1] == self.string[last_char_index]:
                    # prefix is already in tree
                    break
                parent_node_id = self._split_edge(e, self.active)

            node = register_new_node()
            self.nodes[node.id] = node
            edge_id = make_edge_id(parent_node_id, self.string[last_char_index])
            e = register_new_edge(
                edge_id=edge_id,
                first_char_index=last_char_index,
                last_char_index=self.N,
                source_node_id=parent_node_id,
                dest_node_id=node.id,
            )
            self._insert_edge(e)

            if last_parent_node_id is not None:
                self.nodes[last_parent_node_id].suffix_node_id = parent_node_id
            last_parent_node_id = parent_node_id

            if self.active.source_node_id == self.root_node_id:
                self.active.first_char_index += 1
            else:
                self.active.source_node_id = self.nodes[self.active.source_node_id].suffix_node_id
            self._canonize_suffix(self.active)
        if last_parent_node_id is not None:
            self.nodes[last_parent_node_id].suffix_node_id = parent_node_id
        self.active.last_char_index += 1
        self._canonize_suffix(self.active)

    def _insert_edge(self, edge):
        edge_id = make_edge_id(edge.source_node_id, self.string[edge.first_char_index])
        self.edges[edge_id] = edge

    def _remove_edge(self, edge):
        edge_id = make_edge_id(edge.source_node_id, self.string[edge.first_char_index])
        self.edges.pop(edge_id)

    def _split_edge(self, first_edge, suffix):
        assert isinstance(first_edge, Edge)
        assert isinstance(suffix, Suffix)

        # Create a middle node that will split the edge.
        new_node = register_new_node(suffix.source_node_id)
        self.nodes[new_node.id] = new_node

        # Split the char indexes.
        first_edge_last_char_index = first_edge.first_char_index + suffix.length
        second_edge_first_char_index = first_edge.first_char_index + suffix.length + 1
        second_edge_last_char_index = first_edge.last_char_index

        # Create a new edge, from the middle node to the original destination.
        second_edge_id = make_edge_id(new_node.id, self.string[first_edge.first_char_index + suffix.length + 1])
        second_edge = register_new_edge(
            edge_id=second_edge_id,
            first_char_index=second_edge_first_char_index,
            last_char_index=second_edge_last_char_index,
            source_node_id=new_node.id,
            dest_node_id=first_edge.dest_node_id,
        )
        self._insert_edge(second_edge)

        # Shorten the first edge.
        first_edge.last_char_index = first_edge_last_char_index
        first_edge.dest_node_id = new_node.id

        # Return middle node.
        return new_node.id

    def _canonize_suffix(self, suffix):
        """This canonizes the suffix, walking along its suffix string until it
        is explicit or there are no more matched nodes.
        """
        if not suffix.explicit():
            edge_id = make_edge_id(suffix.source_node_id, self.string[suffix.first_char_index])
            e = self.edges[edge_id]
            if e.length <= suffix.length:
                suffix.first_char_index += e.length + 1
                suffix.source_node_id = e.dest_node_id
                self._canonize_suffix(suffix)


class Node(TimestampedVersionedEntity):
    """A node in the suffix tree.
    """

    class Created(Created): pass

    class AttributeChanged(AttributeChanged): pass

    class Discarded(Discarded): pass

    def __init__(self, suffix_node_id=None, *args, **kwargs):
        super(Node, self).__init__(*args, **kwargs)
        self._suffix_node_id = suffix_node_id

    @attribute
    def suffix_node_id(self):
        """The id of a node with a matching suffix, representing a suffix link.

        None indicates this node has no suffix link.
        """
        return self._suffix_node_id

    def __repr__(self):
        return "Node(suffix link: %d)" % self.suffix_node_id


class Edge(TimestampedVersionedEntity):
    """An edge in the suffix tree.
    """

    class Created(Created): pass

    class AttributeChanged(AttributeChanged): pass

    class Discarded(Discarded): pass

    def __init__(self, first_char_index, last_char_index, source_node_id, dest_node_id, **kwargs):
        super(Edge, self).__init__(**kwargs)
        self._first_char_index = first_char_index
        self._last_char_index = last_char_index
        self._source_node_id = source_node_id
        self._dest_node_id = dest_node_id

    @attribute
    def first_char_index(self):
        """Index of start of string part represented by this edge.
        """
        return self._first_char_index

    @attribute
    def last_char_index(self):
        """Index of end of string part represented by this edge.
        """
        return self._last_char_index

    @attribute
    def source_node_id(self):
        """Id of source node of edge.
        """
        return self._source_node_id

    @attribute
    def dest_node_id(self):
        """Id of destination node of edge.
        """
        return self._dest_node_id

    @property
    def length(self):
        """Number of chars in the string part represented by this edge.
        """
        return self.last_char_index - self.first_char_index

    def __repr__(self):
        return 'Edge(%d, %d, %d, %d)' % (self.source_node_id, self.dest_node_id
                                         , self.first_char_index, self.last_char_index)


class Suffix(object):
    """Represents a suffix from first_char_index to last_char_index.
    """

    def __init__(self, source_node_id, first_char_index, last_char_index):
        self._source_node_id = source_node_id
        self._first_char_index = first_char_index
        self._last_char_index = last_char_index

    @property
    def source_node_id(self):
        """Index of node where this suffix starts.
        """
        return self._source_node_id

    @source_node_id.setter
    def source_node_id(self, value):
        self._source_node_id = value

    @property
    def first_char_index(self):
        """Index of start of suffix in string.
        """
        return self._first_char_index

    @first_char_index.setter
    def first_char_index(self, value):
        self._first_char_index = value

    @property
    def last_char_index(self):
        """Index of end of suffix in string.
        """
        return self._last_char_index

    @last_char_index.setter
    def last_char_index(self, value):
        self._last_char_index = value

    @property
    def length(self):
        """Number of chars in string.
        """
        return self.last_char_index - self.first_char_index

    def explicit(self):
        """A suffix is explicit if it ends on a node. first_char_index
        is set greater than last_char_index to indicate this.
        """
        return self.first_char_index > self.last_char_index

    def implicit(self):
        return self.last_char_index >= self.first_char_index


# Factory methods.

def register_new_node(suffix_node_id=None):
    """Factory method, registers new node.
    """
    node_id = uuid4()
    event = Node.Created(originator_id=node_id, suffix_node_id=suffix_node_id)
    entity = Node.mutate(event=event)
    publish(event)
    return entity


def make_edge_id(source_node_index, first_char):
    """Returns a string made from given params.
    """
    return "{}::{}".format(source_node_index, first_char)


def register_new_edge(edge_id, first_char_index, last_char_index, source_node_id, dest_node_id):
    """Factory method, registers new edge.
    """
    event = Edge.Created(
        originator_id=edge_id,
        first_char_index=first_char_index,
        last_char_index=last_char_index,
        source_node_id=source_node_id,
        dest_node_id=dest_node_id,
    )
    entity = Edge.mutate(event=event)
    publish(event)
    return entity


def register_new_suffix_tree(case_insensitive=False):
    """Factory method, returns new suffix tree object.
    """
    assert isinstance(case_insensitive, bool)
    root_node = register_new_node()

    suffix_tree_id = uuid4()
    event = SuffixTree.Created(
        originator_id=suffix_tree_id,
        root_node_id=root_node.id,
        case_insensitive=case_insensitive,
    )
    entity = SuffixTree.mutate(event=event)

    assert isinstance(entity, SuffixTree)

    entity.nodes[root_node.id] = root_node

    publish(event)

    return entity


# Repositories.

class SuffixTreeRepository(AbstractEntityRepository):
    pass


class NodeRepository(AbstractEntityRepository):
    pass


class EdgeRepository(AbstractEntityRepository):
    pass


# Domain services

def find_substring(substring, suffix_tree, edge_repo):
    """Returns the index if substring in tree, otherwise -1.
    """
    assert isinstance(substring, six.string_types)
    assert isinstance(suffix_tree, SuffixTree)
    assert isinstance(edge_repo, EdgeRepository)
    if not substring:
        return -1
    if suffix_tree.case_insensitive:
        substring = substring.lower()
    curr_node_id = suffix_tree.root_node_id
    i = 0
    while i < len(substring):
        edge_id = make_edge_id(curr_node_id, substring[i])
        try:
            edge = edge_repo[edge_id]
        except RepositoryKeyError:
            return -1
        ln = min(edge.length + 1, len(substring) - i)
        if substring[i:i + ln] != suffix_tree.string[edge.first_char_index:edge.first_char_index + ln]:
            return -1
        i += edge.length + 1
        curr_node_id = edge.dest_node_id
    return edge.first_char_index - len(substring) + ln


def has_substring(substring, suffix_tree, edge_repo):
    return find_substring(substring, suffix_tree, edge_repo) != -1


# Application
from eventsourcing.contrib.suffixtrees.infrastructure.respositories.suffixtree_repo import SuffixTreeRepo, \
    NodeRepo, EdgeRepo


class SuffixTreeApplication(ExampleApplication):
    def __init__(self, **kwargs):
        super(SuffixTreeApplication, self).__init__(**kwargs)
        self.suffix_tree_repo = SuffixTreeRepo(self.event_store)
        self.node_repo = NodeRepo(self.event_store)
        self.edge_repo = EdgeRepo(self.event_store)

    def close(self):
        super(SuffixTreeApplication, self).close()
        self.suffix_tree_repo = None
        self.node_repo = None
        self.edge_repo = None

    def register_new_suffixtree(self, case_insensitive=False):
        return register_new_suffix_tree(case_insensitive)

    def find_substring(self, substring, suffix_tree_id):
        suffix_tree = self.suffix_tree_repo[suffix_tree_id]
        started = datetime.datetime.now()
        result = find_substring(substring=substring, suffix_tree=suffix_tree, edge_repo=self.edge_repo)
        print("- found substring '{}' in: {}".format(substring, datetime.datetime.now() - started))
        return result

    def has_substring(self, substring, suffix_tree_id):
        suffix_tree = self.suffix_tree_repo[suffix_tree_id]
        return has_substring(
            substring=substring,
            suffix_tree=suffix_tree,
            edge_repo=self.edge_repo,
        )



from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.datastore import SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager, TrackingRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import IntegerSequencedWithIDRecord, SnapshotRecord, \
    StoredEventRecord, TimestampSequencedNoIDRecord


class SQLAlchemyInfrastructureFactory(InfrastructureFactory):
    record_manager_class = SQLAlchemyRecordManager
    integer_sequenced_record_class = IntegerSequencedWithIDRecord
    timestamp_sequenced_record_class = TimestampSequencedNoIDRecord
    snapshot_record_class = SnapshotRecord
    tracking_record_manager_class = TrackingRecordManager

    def __init__(self, session, uri=None, pool_size=5, *args, **kwargs):
        super(SQLAlchemyInfrastructureFactory, self).__init__(*args, **kwargs)
        self.session = session
        self.uri = uri
        self.pool_size = pool_size

    def construct_record_manager(self, **kwargs):
        s = super(SQLAlchemyInfrastructureFactory, self)
        return s.construct_record_manager(session=self.session, **kwargs)

    def construct_datastore(self):
        datastore = SQLAlchemyDatastore(
            settings=SQLAlchemySettings(
                uri=self.uri,
                pool_size=self.pool_size
            ),
            session=self.session,
        )
        self.session = datastore.session
        return datastore

    def construct_tracking_record_manager(self):
        s = super(SQLAlchemyInfrastructureFactory, self)
        return s.construct_tracking_record_manager(self.session)


def construct_sqlalchemy_eventstore(session,
                                    sequenced_item_class=None,
                                    sequence_id_attr_name=None,
                                    position_attr_name=None,
                                    json_encoder_class=None,
                                    json_decoder_class=None,
                                    cipher=None,
                                    record_class=None,
                                    contiguous_record_ids=False,
                                    application_name=None,
                                    pipeline_id=-1,
                                    ):
    sequenced_item_class = sequenced_item_class or StoredEvent
    sequenced_item_mapper = SequencedItemMapper(
        sequenced_item_class=sequenced_item_class,
        sequence_id_attr_name=sequence_id_attr_name,
        position_attr_name=position_attr_name,
        json_encoder_class=json_encoder_class,
        json_decoder_class=json_decoder_class,
        cipher=cipher,
    )
    factory = SQLAlchemyInfrastructureFactory(
        session=session,
        integer_sequenced_record_class=record_class or StoredEventRecord,
        sequenced_item_class=sequenced_item_class,
        contiguous_record_ids=contiguous_record_ids,
        application_name=application_name,
        pipeline_id=pipeline_id,
    )
    record_manager = factory.construct_integer_sequenced_record_manager()
    event_store = EventStore(
        record_manager=record_manager,
        sequenced_item_mapper=sequenced_item_mapper,
    )
    return event_store



from abc import ABCMeta

from six import with_metaclass


class PipelineExpression(object):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __or__(self, other):
        return PipelineExpression(self, other)

    def __iter__(self):
        for term in (self.left, self.right):
            if isinstance(term, PipelineExpression):
                for item in term:
                    yield item
            else:
                assert issubclass(term, Pipeable), term
                yield term


class PipeableMetaclass(ABCMeta):
    def __or__(self, other):
        return PipelineExpression(self, other)


class Pipeable(with_metaclass(PipeableMetaclass)):
    pass



import datetime

from eventsourcing.example.application import ApplicationWithPersistencePolicies
from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import GeneralizedSuffixTree, \
    register_new_suffix_tree
from eventsourcing.contrib.suffixtrees.domain.services.generalizedsuffixtree import find_substring_edge, \
    get_string_ids, has_substring
from eventsourcing.contrib.suffixtrees.infrastructure.respositories.generalizedsuffixtree_repo import EdgeRepo, \
    GeneralizedSuffixTreeRepo, NodeChildCollectionRepo, NodeRepo, StringidCollectionRepo


class SuffixTreeApplication(ApplicationWithPersistencePolicies):
    def __init__(self, **kwargs):
        super(SuffixTreeApplication, self).__init__(**kwargs)
        self.suffix_tree_repo = GeneralizedSuffixTreeRepo(self.event_store)
        self.node_repo = NodeRepo(self.event_store)
        self.node_child_collection_repo = NodeChildCollectionRepo(self.event_store)
        self.edge_repo = EdgeRepo(self.event_store)
        self.stringid_collection_repo = StringidCollectionRepo(self.event_store)

    def close(self):
        super(SuffixTreeApplication, self).close()
        self.suffix_tree_repo = None
        self.node_repo = None
        self.node_child_collection_repo = None
        self.edge_repo = None
        self.stringid_collection_repo = None

    def register_new_suffix_tree(self, case_insensitive=False):
        """Returns a new suffix tree entity.
        """
        suffix_tree = register_new_suffix_tree(case_insensitive=case_insensitive)
        suffix_tree._node_repo = self.node_repo
        suffix_tree._node_child_collection_repo = self.node_child_collection_repo
        suffix_tree._edge_repo = self.edge_repo
        suffix_tree._stringid_collection_repo = self.stringid_collection_repo
        return suffix_tree

    def get_suffix_tree(self, suffix_tree_id):
        """Returns a suffix tree entity, equipped with node and edge repos it (at least at the moment) needs.
        """
        suffix_tree = self.suffix_tree_repo[suffix_tree_id]
        assert isinstance(suffix_tree, GeneralizedSuffixTree)
        suffix_tree._node_repo = self.node_repo
        suffix_tree._node_child_collection_repo = self.node_child_collection_repo
        suffix_tree._edge_repo = self.edge_repo
        suffix_tree._stringid_collection_repo = self.stringid_collection_repo
        return suffix_tree

    def find_string_ids(self, substring, suffix_tree_id, limit=None):
        """Returns a set of IDs for strings that contain the given substring.
        """

        # Find an edge for the substring.
        edge, ln = self.find_substring_edge(substring=substring, suffix_tree_id=suffix_tree_id)

        # If there isn't an edge, return an empty set.
        if edge is None:
            return set()

        # Get all the string IDs beneath the edge's destination node.
        string_ids = get_string_ids(
            node_id=edge.dest_node_id,
            node_repo=self.node_repo,
            node_child_collection_repo=self.node_child_collection_repo,
            stringid_collection_repo=self.stringid_collection_repo,
            length_until_end=edge.length + 1 - ln,
            limit=limit
        )

        # Return a set of string IDs.
        return set(string_ids)

    def find_substring_edge(self, substring, suffix_tree_id):
        """Returns an edge that matches the given substring.
        """
        suffix_tree = self.suffix_tree_repo[suffix_tree_id]
        started = datetime.datetime.now()
        edge, ln = find_substring_edge(substring=substring, suffix_tree=suffix_tree, edge_repo=self.edge_repo)
        # if edge is not None:
        #     print("Got edge for substring '{}': {}".format(substring, edge))
        # else:
        #     print("No edge for substring '{}'".format(substring))
        print(" - searched for edge in {} for substring: '{}'".format(datetime.datetime.now() - started, substring))
        return edge, ln

    def has_substring(self, substring, suffix_tree_id):
        suffix_tree = self.suffix_tree_repo[suffix_tree_id]
        return has_substring(
            substring=substring,
            suffix_tree=suffix_tree,
            edge_repo=self.edge_repo,
        )



from eventsourcing.domain.model.aggregate import AggregateRoot


class Command(AggregateRoot):
    def __init__(self, **kwargs):
        super(Command, self).__init__(**kwargs)
        self._is_done = False

    class Event(AggregateRoot.Event):
        pass

    class Created(Event, AggregateRoot.Created):
        pass

    @property
    def is_done(self):
        return self._is_done

    def done(self):
        self.__trigger_event__(self.Done)

    class Done(Event):
        def mutate(self, obj):
            obj._is_done = True



import os
import time
from unittest import TestCase
from uuid import uuid4

from eventsourcing.application.multiprocess import Multiprocess
from eventsourcing.application.system import System
from eventsourcing.domain.model.events import assert_event_handlers_empty, clear_event_handlers
from eventsourcing.tests.test_process import ExampleAggregate
from eventsourcing.tests.test_system_fixtures import set_db_uri, Order, Reservation, Payment, create_new_order, Orders, Reservations, Payments, Examples


class TestSystem(TestCase):

    def test_singlethreaded_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True
        )

        with system:
            # Create new Order aggregate.
            order_id = create_new_order()

            # Check the order is reserved and paid.
            repository = system.orders.repository
            assert repository[order_id].is_reserved
            assert repository[order_id].is_paid

    def test_multiprocessing_singleapp_system(self):
        system = System(Examples | Examples, setup_tables=True)

        set_db_uri()

        with Examples() as app, Multiprocess(system):
            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            assert aggregate.id in app.repository

            # Check the aggregate is moved on.
            retries = 50
            while not app.repository[aggregate.id].is_moved_on:

                time.sleep(0.1)
                retries -= 1
                assert retries, "Failed to move"

    def test_multiprocessing_multiapp_system(self):

        set_db_uri()

        with Orders(setup_table=True) as app:
            # Create a new order.
            order_id = create_new_order()

            # Check new order exists in the repository.
            assert order_id in app.repository

        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
        )

        multiprocess = Multiprocess(system)

        # Start multiprocessing system.
        with multiprocess:

            with Orders() as app:

                retries = 50
                while not app.repository[order_id].is_reserved:
                    time.sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved"

                while retries and not app.repository[order_id].is_paid:
                    time.sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid"

    def test_multipipeline_multiprocessing_multiapp(self):

        set_db_uri()

        system = System(
            (Orders, Reservations, Orders, Payments, Orders),
            setup_tables=True
        )

        num_pipelines = 3

        pipeline_ids = range(num_pipelines)

        multiprocess = Multiprocess(system, pipeline_ids=pipeline_ids)

        num_orders_per_pipeline = 5
        order_ids = []

        # Start multiprocessing system.
        with multiprocess, Orders(setup_table=True) as orders:

            # Create some new orders.
            for _ in range(num_orders_per_pipeline):

                for pipeline_id in pipeline_ids:

                    orders.change_pipeline(pipeline_id)

                    order_id = create_new_order()
                    order_ids.append(order_id)

                    time.sleep(0.05)

            # Wait for orders to be reserved and paid.
            retries = 10 + 10 * num_orders_per_pipeline * len(pipeline_ids)
            for i, order_id in enumerate(order_ids):

                while not orders.repository[order_id].is_reserved:
                    time.sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved {} ({})".format(order_id, i)

                while retries and not orders.repository[order_id].is_paid:
                    time.sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid ({})".format(i)

            # Calculate timings from event timestamps.
            order_aggregates = [orders.repository[oid] for oid in order_ids]
            first_timestamp = min([o.__created_on__ for o in order_aggregates])
            last_timestamp = max([o.__last_modified__ for o in order_aggregates])
            duration = last_timestamp - first_timestamp
            rate = len(order_ids) / float(duration)
            period = 1 / rate
            print("Orders system processed {} orders in {:.3f}s at rate of {:.1f} "
                  "orders/s, {:.3f}s each".format(len(order_ids), duration, rate, period))

            # Print min, average, max duration.
            durations = [o.__last_modified__ - o.__created_on__ for o in order_aggregates]
            print("Min order processing time: {:.3f}s".format(min(durations)))
            print("Mean order processing time: {:.3f}s".format(sum(durations) / len(durations)))
            print("Max order processing time: {:.3f}s".format(max(durations)))

    def test_payments_policy(self):
        # Prepare fake repository with a real Order aggregate.
        order = Order.__create__()
        fake_repository = {order.id: order}

        # Check policy makes payment whenever order is reserved.
        event = Order.Reserved(originator_id=order.id, originator_version=1)

        with Payments() as process:
            payment = process.policy(repository=fake_repository, event=event)
            assert isinstance(payment, Payment), payment
            assert payment.order_id == order.id

    def test_orders_policy(self):
        # Prepare fake repository with a real Order aggregate.
        order = Order.__create__()
        fake_repository = {order.id: order}

        # Check order is not reserved.
        assert not order.is_reserved

        # Reservation created.
        with Orders() as process:
            event = Reservation.Created(originator_id=uuid4(), originator_topic='', order_id=order.id)
            process.policy(repository=fake_repository, event=event)

        # Check order is reserved.
        assert order.is_reserved

    def tearDown(self):
        assert_event_handlers_empty()
        clear_event_handlers()
        try:
            del (os.environ['DB_URI'])
        except KeyError:
            pass


# Second example - orders, reservations, payments.





import os

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.interface.notificationlog import RecordManagerNotificationLog
from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.utils.random import decode_random_bytes


class SimpleApplication(object):
    persist_event_type = None
    sequenced_item_class = None
    sequenced_item_mapper_class = None
    infrastructure_factory_class = None
    record_manager_class = None
    stored_event_record_class = None
    snapshot_record_class = None
    json_encoder_class = None
    json_decoder_class = None
    is_constructed_with_session = False

    def __init__(self, name='', persistence_policy=None, persist_event_type=None,
                 cipher_key=None, sequenced_item_class=None, sequenced_item_mapper_class=None,
                 infrastructure_factory_class=None, record_manager_class=None,
                 stored_event_record_class=None, snapshot_record_class=None,
                 setup_table=True, contiguous_record_ids=True, pipeline_id=-1,
                 json_encoder_class=None, json_decoder_class=None,
                 notification_log_section_size=None):

        self.name = name or type(self).__name__.lower()

        self.notification_log_section_size = notification_log_section_size

        self.sequenced_item_class = sequenced_item_class \
                                    or type(self).sequenced_item_class \
                                    or StoredEvent

        self.sequenced_item_mapper_class = sequenced_item_mapper_class \
                                           or type(self).sequenced_item_mapper_class \
                                           or SequencedItemMapper

        self.infrastructure_factory_class = infrastructure_factory_class \
                                            or type(self).infrastructure_factory_class \
                                            or InfrastructureFactory

        self.record_manager_class = record_manager_class or type(self).record_manager_class

        self.stored_event_record_class = stored_event_record_class or type(self).stored_event_record_class

        self.snapshot_record_class = snapshot_record_class or type(self).snapshot_record_class

        self.json_encoder_class = json_encoder_class or type(self).json_encoder_class

        self.json_decoder_class = json_decoder_class or type(self).json_decoder_class

        self.persist_event_type = persist_event_type or type(self).persist_event_type

        self.contiguous_record_ids = contiguous_record_ids
        self.pipeline_id = pipeline_id
        self.setup_cipher(cipher_key)
        self.setup_infrastructure()
        if setup_table:
            self.setup_table()
        self.setup_notification_log()

        self.persistence_policy = persistence_policy
        if self.persistence_policy is None:
            self.setup_persistence_policy()

    def setup_cipher(self, cipher_key):
        cipher_key = decode_random_bytes(cipher_key or os.getenv('CIPHER_KEY', ''))
        self.cipher = AESCipher(cipher_key) if cipher_key else None

    def setup_infrastructure(self, *args, **kwargs):
        self.infrastructure_factory = self.construct_infrastructure_factory(*args, **kwargs)
        self.datastore = self.infrastructure_factory.construct_datastore()
        self.setup_event_store()
        self.setup_repository()

    def construct_infrastructure_factory(self, *args, **kwargs):
        """
        :rtype: InfrastructureFactory
        """
        return self.infrastructure_factory_class(
            record_manager_class=self.record_manager_class,
            integer_sequenced_record_class=self.stored_event_record_class,
            sequenced_item_class=self.sequenced_item_class,
            contiguous_record_ids=self.contiguous_record_ids,
            application_name=self.name,
            pipeline_id=self.pipeline_id,
            snapshot_record_class=self.snapshot_record_class,
            *args, **kwargs
        )

    def setup_event_store(self):
        # Construct event store.
        sequenced_item_mapper = self.sequenced_item_mapper_class(
            sequenced_item_class=self.sequenced_item_class,
            cipher=self.cipher,
            # sequence_id_attr_name=sequence_id_attr_name,
            # position_attr_name=position_attr_name,
            json_encoder_class=self.json_encoder_class,
            json_decoder_class=self.json_decoder_class,
        )
        record_manager = self.infrastructure_factory.construct_integer_sequenced_record_manager()
        self.event_store = EventStore(
            record_manager=record_manager,
            sequenced_item_mapper=sequenced_item_mapper,
        )

    def setup_repository(self, **kwargs):
        self.repository = EventSourcedRepository(
            event_store=self.event_store,
            **kwargs
        )

    def setup_table(self):
        # Setup the database table using event store's record class.
        if self.datastore is not None:
            self.datastore.setup_table(
                self.event_store.record_manager.record_class
            )

    def setup_notification_log(self):
        self.notification_log = RecordManagerNotificationLog(
            self.event_store.record_manager,
            section_size=self.notification_log_section_size
        )

    def setup_persistence_policy(self):
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store,
            event_type=self.persist_event_type
        )

    def change_pipeline(self, pipeline_id):
        self.pipeline_id = pipeline_id
        self.event_store.record_manager.pipeline_id = pipeline_id

    def drop_table(self):
        # Drop the database table using event store's record class.
        if self.datastore is not None:
            self.datastore.drop_table(
                self.event_store.record_manager.record_class
            )

    def close(self):
        # Close the persistence policy.
        if self.persistence_policy is not None:
            self.persistence_policy.close()

        # Close database connection.
        if self.datastore is not None:
            self.datastore.close_connection()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from base64 import b64encode, b64decode

import os


def encode_random_bytes(num_bytes):
    """Generates random bytes, encoded as Base64 unicode string."""
    return b64encode(random_bytes(num_bytes)).decode('utf-8')


def random_bytes(num_bytes):
    return os.urandom(num_bytes)


def decode_random_bytes(s):
    """Returns bytes, decoded from Base64 encoded unicode string."""
    return b64decode(s.encode('utf-8'))



import os
import unittest

import django
os.environ['DJANGO_SETTINGS_MODULE'] = 'eventsourcing.tests.djangoproject.djangoproject.settings'
django.setup()

from django.core.management import call_command

from eventsourcing.infrastructure.django.apps import DjangoConfig
from eventsourcing.infrastructure.django.factory import DjangoInfrastructureFactory

from django.test import TransactionTestCase

from eventsourcing.tests.sequenced_item_tests.base import IntegerSequencedRecordTestCase, TimestampSequencedItemTestCase


class DjangoTestCase(TransactionTestCase):

    infrastructure_factory_class = DjangoInfrastructureFactory
    contiguous_record_ids = True

    def setUp(self):
        super(DjangoTestCase, self).setUp()
        call_command('migrate', verbosity=0, interactive=False)

    def construct_datastore(self):
        pass


class TestDjangoRecordManagerWithIntegerSequences(DjangoTestCase, IntegerSequencedRecordTestCase):
    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestDjangoRecordManagerWithoutContiguousRecordIDs(DjangoTestCase, IntegerSequencedRecordTestCase):
    contiguous_record_ids = False
    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestDjangoRecordManagerWithTimestampSequences(DjangoTestCase, TimestampSequencedItemTestCase):
    def construct_record_manager(self):
        return self.construct_timestamp_sequenced_record_manager()


# class WithDjangoRecordManagers(DjangoTestCase, WithActiveRecordManagers):
#     def construct_entity_record_manager(self):
#         return construct_integer_sequenced_record_manager()
#
#     def construct_log_record_manager(self):
#         return construct_timestamp_sequenced_record_manager()
#
#     def construct_snapshot_record_manager(self):
#         return construct_snapshot_record_manager()
#

# class TestSimpleSequencedItemIteratorWithDjango(WithDjangoRecordManagers,
#                                                 SimpleSequencedItemteratorTestCase):
#     pass
#
#
# class TestThreadedSequencedItemIteratorWithDjango(WithDjangoRecordManagers,
#                                                   ThreadedSequencedItemIteratorTestCase):
#     pass

class TestConfigClass(unittest.TestCase):

    def test(self):
        self.assertEqual('eventsourcing.infrastructure.django', DjangoConfig.name)



from unittest import TestCase

from eventsourcing.exceptions import DataIntegrityError


class TestAESCipher(TestCase):

    def test_encrypt_mode_gcm(self):
        from eventsourcing.utils.cipher.aes import AESCipher
        from eventsourcing.utils.random import encode_random_bytes, decode_random_bytes

        # Unicode string representing 256 random bits encoded with Base64.
        cipher_key = encode_random_bytes(num_bytes=32)

        # Construct AES cipher.
        cipher = AESCipher(cipher_key=decode_random_bytes(cipher_key))

        # Encrypt some plaintext.
        ciphertext = cipher.encrypt('plaintext')
        self.assertNotEqual(ciphertext, 'plaintext')

        # Decrypt some ciphertext.
        plaintext = cipher.decrypt(ciphertext)
        self.assertEqual(plaintext, 'plaintext')

        # Check DataIntegrityError is raised (broken Base64 padding).
        with self.assertRaises(DataIntegrityError):
            damaged = ciphertext[:-1]
            cipher.decrypt(damaged)

        # Check DataIntegrityError is raised (MAC check fails).
        with self.assertRaises(DataIntegrityError):
            damaged = 'a' + ciphertext[:-1]
            cipher.decrypt(damaged)

        # Check DataIntegrityError is raised (nonce too short).
        with self.assertRaises(DataIntegrityError):
            damaged = ciphertext[:0]
            cipher.decrypt(damaged)

        # Check DataIntegrityError is raised (tag too short).
        with self.assertRaises(DataIntegrityError):
            damaged = ciphertext[:20]
            cipher.decrypt(damaged)



class EventSourcingError(Exception):
    """Base eventsourcing exception."""


class TopicResolutionError(EventSourcingError):
    """Raised when unable to resolve a topic to a Python class."""


class EntityVersionNotFound(EventSourcingError):
    """Raise when accessing an entity version that does not exist."""


class RecordConflictError(EventSourcingError):
    """Raised when database raises an integrity error."""


class PromptFailed(EventSourcingError):
    """Raised when prompt fails."""


class ConcurrencyError(RecordConflictError):
    """Raised when a record conflict is due to concurrency."""


class ConsistencyError(EventSourcingError):
    """Raised when applying an event stream to a versioned entity."""


class MismatchedOriginatorError(ConsistencyError):
    """Raised when applying an event to an inappropriate object."""


class OriginatorIDError(MismatchedOriginatorError):
    """Raised when applying an event to the wrong entity or aggregate."""


class OriginatorVersionError(MismatchedOriginatorError):
    """Raised when applying an event to the wrong version of an entity or aggregate."""


class MutatorRequiresTypeNotInstance(ConsistencyError):
    """Raised when mutator function received a class rather than an entity."""


class DataIntegrityError(ValueError, EventSourcingError):
    "Raised when a sequenced item data is damaged (hash doesn't match data)"


class EventHashError(DataIntegrityError):
    "Raised when an event's seal hash doesn't match the hash of the state of the event."


class HeadHashError(DataIntegrityError, MismatchedOriginatorError):
    """Raised when applying an event with hash different from aggregate head."""


class EntityIsDiscarded(AssertionError):
    """Raised when access to a recently discarded entity object is attempted."""


class ProgrammingError(EventSourcingError):
    """Raised when programming errors are encountered."""


class RepositoryKeyError(KeyError, EventSourcingError):
    """Raised when using entity repository's dictionary like interface  to get an entity that does not exist."""


class ArrayIndexError(IndexError, EventSourcingError):
    """Raised when appending item to an array that is full."""


class DatasourceSettingsError(EventSourcingError):
    "Raised when an error is detected in settings for a datasource."


class OperationalError(EventSourcingError):
    "Raised when an operational error is encountered."


class TimeSequenceError(EventSourcingError):
    "Raised when a time sequence error occurs e.g. trying to save a timestamp that already exists."

class TrackingRecordNotFound(EventSourcingError):
    "Raised when a tracking record is not found."

class CausalDependencyFailed(EventSourcingError):
    "Raised when a causal dependency fails (after its tracking record not found)."

class EventRecordNotFound(EventSourcingError):
    "Raised when an event record is not found."



import time
from datetime import datetime, timedelta
from unittest.case import TestCase
from uuid import uuid1

import sys

from eventsourcing.utils.random import encode_random_bytes, decode_random_bytes
from eventsourcing.utils.times import decimaltimestamp_from_uuid, utc_timezone, decimaltimestamp


class TestUtils(TestCase):
    def test_timestamp_from_uuid(self):
        until = decimaltimestamp()
        uuid = uuid1()
        after = decimaltimestamp()
        uuid_timestamp = decimaltimestamp_from_uuid(uuid)
        self.assertLess(until, uuid_timestamp)
        self.assertGreater(after, uuid_timestamp)

        # Check timestamp_from_uuid() works with hex strings, as well as UUID objects.
        self.assertEqual(decimaltimestamp_from_uuid(uuid.hex), decimaltimestamp_from_uuid(uuid))

    def test_utc(self):
        now = datetime.now(tz=utc_timezone)
        self.assertEqual(utc_timezone.utcoffset(now), timedelta(0))
        expected_dst = None if int(sys.version[0]) > 2 else timedelta(0)
        self.assertEqual(utc_timezone.dst(now), expected_dst)

    def test_encode_random_bytes(self):
        key = encode_random_bytes(num_bytes=16)
        self.assertEqual(len(decode_random_bytes(key)), 16)

        key = encode_random_bytes(num_bytes=24)
        self.assertEqual(len(decode_random_bytes(key)), 24)

        key = encode_random_bytes(num_bytes=32)
        self.assertEqual(len(decode_random_bytes(key)), 32)



from uuid import uuid4

from eventsourcing.example.domainmodel import Example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.test_sqlalchemy import SQLAlchemyDatastoreTestCase
from eventsourcing.utils.topic import get_topic


class TestEventSourcedRepository(SQLAlchemyDatastoreTestCase):
    def setUp(self):
        super(TestEventSourcedRepository, self).setUp()
        if self.datastore is not None:
            self.datastore.setup_connection()
            self.datastore.setup_tables()

    def tearDown(self):
        if self.datastore is not None:
            self.datastore.drop_tables()
            self.datastore.close_connection()
        super(TestEventSourcedRepository, self).tearDown()

    def construct_event_store(self):
        event_store = EventStore(
            record_manager=self.factory.construct_integer_sequenced_record_manager(),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name='originator_id',
                position_attr_name='originator_version',
            )
        )
        return event_store

    def test_get_item(self):
        # Setup an event store.
        event_store = self.construct_event_store()

        # Put an event in the event store.
        entity_id = uuid4()
        event_store.append(Example.Created(
            a=1,
            b=2,
            originator_id=entity_id,
            originator_topic=get_topic(Example),
        ))

        # Construct a repository.
        event_sourced_repo = EventSourcedRepository(event_store=event_store)

        # Check the entity attributes.
        example = event_sourced_repo[entity_id]
        self.assertEqual(1, example.a)
        self.assertEqual(2, example.b)
        self.assertEqual(entity_id, example.id)

        # Setup an example repository, using the subclass ExampleRepository.
        example_repo = ExampleRepository(event_store=event_store)

        # Check the repo has the example.
        self.assertIn(entity_id, example_repo)
        self.assertNotIn(uuid4(), example_repo)

        # Check the entity attributes.
        example = example_repo[entity_id]
        self.assertEqual(1, example.a)
        self.assertEqual(2, example.b)
        self.assertEqual(entity_id, example.id)



from eventsourcing.tests.datastore_tests.test_sqlalchemy import SQLAlchemyDatastoreTestCase
from eventsourcing.tests.sequenced_item_tests.base import IntegerSequencedRecordTestCase, \
    SimpleSequencedItemteratorTestCase, ThreadedSequencedItemIteratorTestCase, TimestampSequencedItemTestCase, \
    WithRecordManagers


class WithSQLAlchemyRecordManagers(SQLAlchemyDatastoreTestCase, WithRecordManagers):
    pass


class TestSQLAlchemyRecordManagerWithIntegerSequences(WithSQLAlchemyRecordManagers,
                                                      IntegerSequencedRecordTestCase):
    pass


class TestSQLAlchemyRecordManagerWithTimestampSequences(WithSQLAlchemyRecordManagers,
                                                        TimestampSequencedItemTestCase):
    pass


class TestSimpleIteratorWithSQLAlchemy(WithSQLAlchemyRecordManagers,
                                       SimpleSequencedItemteratorTestCase):
    pass


class TestThreadedIteratorWithSQLAlchemy(WithSQLAlchemyRecordManagers,
                                         ThreadedSequencedItemIteratorTestCase):
    use_named_temporary_file = True



from eventsourcing.tests.example_application_tests.base import ExampleApplicationTestCase
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import \
    DjangoTestCase


class TestExampleApplicationWithDjango(DjangoTestCase, ExampleApplicationTestCase):
    def construct_log_record_manager(self):
        return self.construct_entity_record_manager()

    def construct_entity_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()




from abc import abstractproperty
from math import ceil, log
from uuid import uuid5

import six

from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.entity import AbstractEntityRepository, TimestampedVersionedEntity
from eventsourcing.domain.model.events import publish
from eventsourcing.exceptions import ConcurrencyError, ArrayIndexError


class ItemAssigned(TimestampedVersionedEntity.Event):
    """Occurs when an item is set at a position in an array."""

    def __init__(self, item, index, *args, **kwargs):
        kwargs['item'] = item
        super(ItemAssigned, self).__init__(originator_version=index, *args, **kwargs)

    @property
    def item(self):
        return self.__dict__['item']

    @property
    def index(self):
        return self.originator_version


class Array(object):
    def __init__(self, array_id, repo):
        assert isinstance(repo, AbstractArrayRepository)
        self.id = array_id
        self.repo = repo

    @retry(ConcurrencyError, max_attempts=50, wait=0.01)
    def append(self, item):
        """Sets item in next position after the last item."""
        self.__setitem__(self.get_next_position(), item)

    def __setitem__(self, index, item):
        """
        Sets item in array, at given index.

        Won't overrun the end of the array, because
        the position is fixed to be less than base_size.
        """
        size = self.repo.array_size
        if size and index >= size:
            raise ArrayIndexError("Index is {}, but size is {}".format(index, size))
        event = ItemAssigned(
            originator_id=self.id,
            index=index,
            item=item,
        )
        publish(event)

    def __getitem__(self, item):
        """
        Returns item at index, or items in slice.
        """
        assert isinstance(item, (six.integer_types, slice))
        array_size = self.repo.array_size
        if isinstance(item, six.integer_types):
            if item < 0:
                index = array_size + item
                if index < 0:
                    raise ArrayIndexError("Array index out of range: {}".format(item))
            else:
                index = item
            event = self.get_item_assigned(index)
            assert isinstance(event, ItemAssigned)
            return event.item
        elif isinstance(item, slice):
            assert item.step in (None, 1), "Slice stepping not supported, yet"
            if item.start is None:
                start_index = 0
            elif item.start < 0:
                start_index = max(array_size + item.start, 0)
            else:
                start_index = item.start

            if not isinstance(item.stop, six.integer_types):
                stop_index = array_size
            elif item.stop < 0:
                stop_index = array_size + item.stop
            else:
                stop_index = item.stop

            if stop_index is not None:
                if stop_index <= 0:
                    return []
                # Avoid passing massive integers into the database.
                if array_size is not None:
                    stop_index = min(stop_index, self.repo.array_size)

            items_assigned = self.get_items_assigned(stop_index=stop_index, start_index=start_index)
            items_dict = {str(i.originator_version): i.item for i in items_assigned}
            items = [items_dict.get(str(i)) for i in range(start_index, stop_index)]
            return items

    def __len__(self):
        """
        Returns length of array.
        """
        return self.repo.array_size

    def get_next_position(self):
        return self.get_last_item_and_next_position()[1]

    def get_last_item_and_next_position(self):
        items_assigned = self.get_items_assigned(limit=1, is_ascending=False)
        if len(items_assigned) == 0:
            return None, 0
        item_assigned = items_assigned[0]
        return item_assigned.item, item_assigned.index + 1

    def get_items_assigned(self, start_index=None, stop_index=None, limit=None, is_ascending=True):
        return self.repo.event_store.get_domain_events(
            originator_id=self.id,
            gte=start_index,
            lt=stop_index,
            limit=limit,
            is_ascending=is_ascending,
        )

    def get_item_assigned(self, index):
        try:
            item_assigned = self.repo.event_store.get_domain_event(
                originator_id=self.id,
                position=index,
            )
        except IndexError as e:
            raise ArrayIndexError(e)
        else:
            return  item_assigned

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)


class BigArray(Array):
    """
    A virtual array holding items in indexed
    positions, across a number of Array instances.

    Getting and setting items at index position is
    supported. Slices are supported, and operate
    across the underlying arrays. Appending is also
    supported.

    BigArray is designed to overcome the concern of
    needing a single large sequence that may not be
    suitably stored in any single partiton. In simple
    terms, if events of an aggregate can fit in a
    partition, we can use the same size partition
    to make a tree of arrays that will certainly
    be capable of sequencing all the events of the
    application in a single stream.

    With normal size base arrays, enterprise applications
    can expect read and write time to be approximately
    constant with respect to the number of items in the array.

    The array is composed of a tree of arrays,
    which gives the capacity equal to the size
    of each array to the power of the size of
    each array. If the arrays are limited to be
    about the maximum size of an aggregate event
    stream (a large number but not too many that
    would cause there to be too much data in any
    one partition, let's say 1000s to be safe)
    then it would be possible to fit such a large
    number of aggregates in the corresponding
    BigArray, that we can be confident it would
    be full.

    Write access time in the worst case, and the time
    to identify the index of the last item in the big
    array, is proportional to the log of the highest
    assigned index to base the underlying array
    size. Write time on average, and read time given an
    index, is constant with respect to the number of items
    in a BigArray.

    Items can be appended in log time in a single thread.
    However, the time between reading the current last index
    and claiming the next position leads to contention and
    retries when there are lots of threads of execution all
    attempting to append items, which inherently limits throughput.

    Todo: Not possible in Cassandra, but maybe do it in a
    transaction in SQLAlchemy?

    An alternative to reading the last item before writing
    the next is to use an integer sequence generator to
    generate a stream of integers. Items can be assigned
    to index positions in a big array, according to the integers
    that are issued. Throughput will then be much better, and
    will be limited only by the rate at which the database can have
    events written to it (unless the number generator is quite slow).

    An external integer sequence generator, such as Redis' INCR
    command, or an auto-incrementing database column, may
    constitute a single point of failure.

    """
    def __init__(self, array_id, repo):
        assert isinstance(repo, AbstractArrayRepository), type(repo)
        super(BigArray, self).__init__(array_id=array_id, repo=repo)

    def get_last_array(self):
        """
        Returns last array in compound.

        :rtype: CompoundSequenceReader
        """
        # Get the root array (might not have been registered).
        root = self.repo[self.id]

        # Get length and last item in the root array.
        apex_id, apex_height = root.get_last_item_and_next_position()

        # Bail if there isn't anything yet.
        if apex_id is None:
            return None, None

        # Get the current apex array.
        apex = self.repo[apex_id]
        assert isinstance(apex, Array)

        # Descend until hitting the bottom.
        array = apex
        array_i = 0
        height = apex_height
        while height > 1:
            height -= 1
            array_id, width = array.get_last_item_and_next_position()
            assert width > 0
            offset = width - 1
            array_i += offset * self.repo.array_size ** height
            array = self.repo[array_id]

        return array, array_i

    def get_last_item_and_next_position(self):
        sequence, i = self.get_last_array()
        if sequence is None:
            return None, 0
        item, n = sequence.get_last_item_and_next_position()
        return item, i + n

    def __getitem__(self, item):
        assert isinstance(item, (six.integer_types, slice))
        # Calculate the i and j of the containing base sequence.
        if isinstance(item, six.integer_types):
            return self.get_item(item)
        elif isinstance(item, slice):
            assert item.step in (None, 1), "Slice step must be 1: {}".format(str(item.step))
            return self.get_slice(item.start, item.stop)

    def get_item(self, position):
        if position < 0:
            last, length = self.get_last_item_and_next_position()
            if position == -1:
                return last
            position = position + length
        size = self.repo.array_size
        n = position // size
        i = n * size
        j = i + size
        offset = position - i
        array_id = self.create_array_id(i, j)
        base_array = self.repo[array_id]
        return base_array[offset]

    def get_slice(self, start, stop):
        array_len = self.repo.array_size ** self.repo.array_size

        if start is None:
            start = 0
        elif start < 0:
            start = max(array_len + start, 0)

        if stop is None:
            stop = array_len
        elif stop < 0:
            stop = array_len + stop

        stop = min(stop, array_len)

        size = self.repo.array_size
        while start < stop:
            n = start // size
            i = n * size
            j = i + size
            substart = start - i
            substop = stop - i
            array_id = self.create_array_id(i, j)
            sequence = self.repo[array_id]
            for item in sequence[substart:substop]:
                yield item
            start = j

    def __setitem__(self, position, item):
        # Calculate start and stop position
        # of the containing base array.
        size = self.repo.array_size
        start = (position // size) * size
        stop = start + size

        # Set the item in the base array.
        array_id = self.create_array_id(start, stop)
        array = self.repo[array_id]
        index = position - start
        array[index] = item

        # Calculate the height of the apex of
        # the compound containing given position
        # (zero-based index in big array).
        required_height = self.calc_required_height(position, size)

        # Set array IDs in containing arrays,
        # up to the required height.
        height = 1
        while height < required_height:
            # Set ID of array in its parent array.
            start, stop, height, index_of_child = self.calc_parent(start, stop, height)
            child_id = array_id
            array_id = self.create_array_id(start, stop)
            array = self.repo[array_id]
            try:
                array[index_of_child] = child_id
            except ConcurrencyError:
                return

        # Set the apex to the root array.
        assert array_id
        root = self.repo[self.id]
        try:
            root[required_height - 1] = array_id
        except ConcurrencyError:
            return

    def __len__(self):
        """
        Returns length of array.
        """
        return self.repo.array_size ** self.repo.array_size

    def calc_required_height(self, n, size):
        if size <= 0:
            raise ValueError("Size must be greater than 0")
        capacity = size ** size
        if n + 1 > capacity:
            raise ArrayIndexError("Contents can't be greater than capacity")
        if n == 0 or n == 1:
            return 1
        min_capacity = max(int(size), int(n + 1))
        required_height = log(min_capacity, int(size))
        # Quash numerical error before calling ceil.
        # - this is occasionally required, for example
        #   with base size of 1000 numerical error
        #   makes the required height one greater
        #   than the correct value, which causes
        #   an index error when assigning apex
        #   ID to the root sequence.
        required_height = round(required_height, 10)
        return int(ceil(required_height))

    def calc_parent(self, i, j, h):
        """
        Returns get_big_array and end of span of parent sequence that contains given child.
        """
        N = self.repo.array_size
        c_i = i
        c_j = j
        c_h = h
        # Calculate the number of the sequence in its row (sequences
        # with same height), from left to right, starting from 0.
        c_n = c_i // (N ** c_h)
        p_n = c_n // N
        # Position of the child ID in the parent array.
        p_p = c_n % N
        # Parent height is child height plus one.
        p_h = c_h + 1
        # Span of sequences in parent row is max size N, to the power of the height.
        span = N ** p_h
        # Calculate parent i and j.
        p_i = p_n * span
        p_j = p_i + span
        # Check the parent i,j bounds the child i,j, ie child span is contained by parent span.
        assert p_i <= c_i, 'i greater on parent than child: {}'.format(p_i, p_j)
        assert p_j >= c_j, 'j less on parent than child: {}'.format(p_i, p_j)
        # Return parent i, j, h, p.
        return p_i, p_j, p_h, p_p

    def create_array_id(self, i, j):
        return uuid5(self.id, str((i, j)))


class AbstractArrayRepository(AbstractEntityRepository):
    """
    Repository for sequence objects.
    """

    def __init__(self, array_size=10000, *args, **kwargs):
        super(AbstractArrayRepository, self).__init__(*args, **kwargs)
        self.array_size = array_size

    def __getitem__(self, array_id):
        """
        Returns sequence for given ID.
        """
        return Array(array_id=array_id, repo=self)


class AbstractBigArrayRepository(AbstractEntityRepository):
    """
    Repository for compound sequence objects.
    """

    @abstractproperty
    def subrepo(self):
        """Sub-sequence repository."""

    def __getitem__(self, array_id):
        """
        Returns sequence for given ID.
        """
        return BigArray(array_id=array_id, repo=self.subrepo)



from functools import reduce

from eventsourcing.domain.model.entity import AbstractEventPlayer
from eventsourcing.infrastructure.eventstore import AbstractEventStore


class EventPlayer(AbstractEventPlayer):
    # The page size by which events are retrieved. If this
    # value is set to a positive integer, the events of
    # the entity will be retrieved in pages, using a series
    # of queries, rather than with one potentially large query.
    __page_size__ = None

    def __init__(self, event_store, snapshot_strategy=None, use_cache=False, mutator_func=None):
        super(EventPlayer, self).__init__()
        # Check we got an event store.
        assert isinstance(event_store, AbstractEventStore), type(event_store)
        self._event_store = event_store
        self._snapshot_strategy = snapshot_strategy
        self._cache = {}
        self._use_cache = use_cache
        self._mutator_func = mutator_func

    @property
    def event_store(self):
        return self._event_store

    def project_events(self, initial_state, domain_events):
        """
        Evolves initial state using the sequence of domain events and a mutator function.
        """
        return reduce(self._mutator_func or self.mutate, domain_events, initial_state)

    @staticmethod
    def mutate(initial, event):
        if initial is not None:
            event.__check_obj__(initial)
        return event.__mutate__(initial)

# def clone_object(initial_state):
#     initial_state_copy = object.__new__(type(initial_state))
#     initial_state_copy.__dict__.update(deepcopy(initial_state.__dict__))
#     return initial_state_copy



from __future__ import unicode_literals

from abc import ABCMeta, abstractmethod

import six

from eventsourcing.infrastructure.sequenceditem import SequencedItem, SequencedItemFieldNames
from eventsourcing.utils.topic import get_topic, resolve_topic
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder, json_dumps, json_loads


class AbstractSequencedItemMapper(six.with_metaclass(ABCMeta)):
    @abstractmethod
    def to_sequenced_item(self, domain_event):
        """
        Returns sequenced item for given domain event.
        """

    @abstractmethod
    def from_sequenced_item(self, sequenced_item):
        """
        Return domain event from given sequenced item.
        """


class SequencedItemMapper(AbstractSequencedItemMapper):
    """
    Uses JSON to transcode domain events.
    """

    def __init__(self, sequenced_item_class=SequencedItem, sequence_id_attr_name=None,
                 position_attr_name=None, json_encoder_class=None, json_decoder_class=None,
                 cipher=None, other_attr_names=()):
        self.sequenced_item_class = sequenced_item_class
        self.json_encoder_class = json_encoder_class or ObjectJSONEncoder
        self.json_decoder_class = json_decoder_class or ObjectJSONDecoder
        self.cipher = cipher
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)
        self.sequence_id_attr_name = sequence_id_attr_name or self.field_names.sequence_id
        self.position_attr_name = position_attr_name or self.field_names.position
        self.other_attr_names = other_attr_names or self.field_names.other_names

    def to_sequenced_item(self, domain_event):
        """
        Constructs a sequenced item from a domain event.
        """
        item_args = self.construct_item_args(domain_event)
        return self.construct_sequenced_item(item_args)

    def construct_item_args(self, domain_event):
        """
        Constructs attributes of a sequenced item from the given domain event.
        """
        # Copy the state of the event.
        event_attrs = domain_event.__dict__.copy()

        # Get the sequence ID.
        sequence_id = event_attrs.get(self.sequence_id_attr_name)

        # Get the position in the sequence.
        position = event_attrs.get(self.position_attr_name)

        # Get the topic from the event attrs, otherwise from the class.
        topic = get_topic(domain_event.__class__)

        # Serialise the remaining event attribute values.
        data = json_dumps(event_attrs, cls=self.json_encoder_class)

        # Encrypt data.
        if self.cipher:
            data = self.cipher.encrypt(data)

        # Get the 'other' args.
        # - these are meant to be derivative of the other attributes,
        #   to populate database fields, and shouldn't affect the hash.
        other_args = tuple((getattr(domain_event, name) for name in self.other_attr_names))

        return (sequence_id, position, topic, data) + other_args

    def construct_sequenced_item(self, item_args):
        return self.sequenced_item_class(*item_args)

    def from_sequenced_item(self, sequenced_item):
        """
        Reconstructs domain event from stored event topic and
        event attrs. Used in the event store when getting domain events.
        """
        assert isinstance(sequenced_item, self.sequenced_item_class), (
            self.sequenced_item_class, type(sequenced_item)
        )

        # Get the topic and data.
        topic = getattr(sequenced_item, self.field_names.topic)
        data = getattr(sequenced_item, self.field_names.data)

        return self.from_topic_and_data(topic, data)

    def from_topic_and_data(self, topic, data):
        # Resolve topic to event class.
        domain_event_class = resolve_topic(topic)

        # Decrypt data.
        if self.cipher:
            data = self.cipher.decrypt(data)

        # Deserialize data.
        event_attrs = json_loads(data, cls=self.json_decoder_class)

        # Reconstruct domain event object.
        return reconstruct_object(domain_event_class, event_attrs)


def reconstruct_object(obj_class, obj_state):
    obj = object.__new__(obj_class)
    obj.__dict__.update(obj_state)
    return obj



import eventsourcing.example.interface.flaskapp


# Todo: Investigate why removing this file breaks python3.3.
# For some reason, this file is needed to run flaskapp.py
# with uWSGI and python3.3.
#
application = eventsourcing.example.interface.flaskapp.application



from math import ceil
from threading import Thread
from uuid import uuid4

from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.infrastructure.repositories.array import BigArrayRepository
from eventsourcing.interface.notificationlog import BigArrayNotificationLog, NotificationLogReader, \
    NotificationLogView, RecordManagerNotificationLog, RemoteNotificationLog
from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicies
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import \
    WithCassandraRecordManagers
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import \
    WithSQLAlchemyRecordManagers
from eventsourcing.utils.topic import get_topic


class NotificationLogTestCase(WithSQLAlchemyRecordManagers, WithPersistencePolicies):

    def assert_section(self, repo, requested_id, expected_id, expected_len_items, expected_previous_id,
                       expected_next_id):
        section = repo[requested_id]
        self.assertEqual(expected_len_items, len(list(section.items)))
        self.assertEqual(expected_id, section.section_id)
        self.assertEqual(expected_previous_id, section.previous_id)
        self.assertEqual(expected_next_id, section.next_id)

    def append_notifications(self, *range_args):
        for i in range(*range_args):
            item = 'item{}'.format(i + 1)
            self.append_notification(item)

    def create_notification_log(self, section_size):
        return RecordManagerNotificationLog(self.entity_record_manager, section_size)

    def construct_entity_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()

    def append_notification(self, item):
        sequenced_item = self.entity_record_manager.sequenced_item_class(
            uuid4(),
            0,
            get_topic(DomainEvent),
            item
        )
        self.entity_record_manager.append(sequenced_item)


class TestNotificationLog(NotificationLogTestCase):

    def test(self):
        # Build notification log.
        section_size = 5
        notification_log = self.create_notification_log(section_size=section_size)

        # Check the sections.
        section = notification_log['current']
        self.assertEqual('1,5', section.section_id)
        self.assertEqual(0, len(list(section.items)))
        self.assertIsNone(section.previous_id)
        self.assertIsNone(section.previous_id)

        # Append notifications.
        self.append_notifications(13)

        # Check the sections.
        self.assert_section(notification_log, 'current', '11,15', 3, '6,10', None)
        self.assert_section(notification_log, '1,5', '1,5', section_size, None, '6,10')
        self.assert_section(notification_log, '6,10', '6,10', section_size, '1,5', '11,15')
        self.assert_section(notification_log, '11,15', '11,15', 3, '6,10', None)
        self.assert_section(notification_log, '16,20', '16,20', 0, '11,15', None)
        self.assert_section(notification_log, '21,25', '21,25', 0, '16,20', None)

        # Add some more notification.
        self.append_notifications(13, 24)

        # Check the notification log has been extended.
        self.assertEqual(len(list(notification_log['11,15'].items)), section_size)
        self.assertEqual(len(list(notification_log['16,20'].items)), section_size)
        self.assertEqual(len(list(notification_log['21,25'].items)), 4)
        self.assertEqual(len(list(notification_log['26,30'].items)), 0)

        # Check an section ID that can't be split by ',' results in a ValueError.
        with self.assertRaises(ValueError):
            _ = notification_log['invalid']


class TestBigArrayNotificationLog(TestNotificationLog):

    def test(self):
        self.section_size = 5
        super(TestBigArrayNotificationLog, self).test()

        # Check array size must be divisible by section size.
        with self.assertRaises(ValueError):
            BigArrayNotificationLog(self.big_array, section_size=6)

        # # Check the section ID must match the section size.
        # notification_log = BigArrayNotificationLog(self.big_array, self.section_size)
        # with self.assertRaises(ValueError):
        #     _ = notification_log['1,2']

        # # Check the section ID must be aligned to the array size.
        # with self.assertRaises(ValueError):
        #     _ = notification_log['2,6']

    def create_notification_log(self, section_size):
        self.big_array = self.create_big_array()
        return BigArrayNotificationLog(self.big_array, section_size=section_size)

    def create_big_array(self, big_array_id=None):
        big_array_id = uuid4() if big_array_id is None else big_array_id
        big_array_repo = BigArrayRepository(event_store=self.entity_event_store)
        big_array = big_array_repo[big_array_id]
        return big_array

    def append_notification(self, item):
        self.big_array.append(item)


class TestNotificationLogReader(NotificationLogTestCase):

    def test(self):
        # Build notification log.
        section_size = 5
        notification_log = self.create_notification_log(section_size=section_size)

        # Append 13 notifications.
        self.append_notifications(13)

        # Construct notification log reader.
        reader = NotificationLogReader(notification_log)

        # Check position.
        self.assertEqual(reader.position, 0)

        # Read all notifications.
        all_notifications = list(reader)
        self.assertEqual(13, len(all_notifications))

        # Check position.
        self.assertEqual(reader.position, 13)

        # Add some more items to the log.
        self.append_notifications(13, 21)

        # Read subsequent notifications.
        subsequent_notifications_notifications = list(reader)
        self.assertEqual(len(subsequent_notifications_notifications), 8)

        # Check position.
        self.assertEqual(reader.position, 21)

        subsequent_notifications_notifications = list(reader)
        self.assertEqual(len(subsequent_notifications_notifications), 0)

        # Set position.
        reader.seek(13)
        subsequent_notifications_notifications = list(reader)
        self.assertEqual(len(subsequent_notifications_notifications), 8)

        # # Read items after a particular position.
        self.assertEqual(len(list(reader[0:])), 21)
        self.assertEqual(len(list(reader[1:])), 20)
        self.assertEqual(len(list(reader[2:])), 19)
        self.assertEqual(len(list(reader[3:])), 18)
        self.assertEqual(len(list(reader[13:])), 8)
        self.assertEqual(len(list(reader[18:])), 3)
        self.assertEqual(len(list(reader[19:])), 2)
        self.assertEqual(len(list(reader[20:])), 1)
        self.assertEqual(len(list(reader[21:])), 0)

        # Check last item numbers less than 1 cause a value errors.
        with self.assertRaises(ValueError):
            reader.position = -1
            list(reader)

        with self.assertRaises(ValueError):
            list(reader.seek(-1))

        # Resume from a saved position.
        saved_position = 5
        advance_by = 3
        reader.seek(saved_position)
        self.assertEqual(reader.position, saved_position)
        reader.read_list(advance_by=advance_by)
        self.assertEqual(reader.position, saved_position + advance_by)

        # Read items between particular positions.
        # - check stops at end of slice, and position tracks ok
        self.assertEqual(reader[0]['id'], 1)
        self.assertEqual(reader.position, 1)

        self.assertEqual(next(reader)['id'], 2)
        self.assertEqual(reader.position, 2)

        reader.seek(5)
        self.assertEqual(next(reader)['id'], 6)
        self.assertEqual(reader.position, 6)

        reader.seek(0)
        list(reader)
        self.assertEqual(reader.position, 21)

        self.assertEqual(len(list(reader[0:1])), 1)
        self.assertEqual(reader.position, 1)

        self.assertEqual(len(list(reader[1:3])), 2)
        self.assertEqual(reader.position, 3)

        self.assertEqual(len(list(reader[2:5])), 3)
        self.assertEqual(reader.position, 5)

        self.assertEqual(len(list(reader[3:7])), 4)
        self.assertEqual(reader.position, 7)

        self.assertEqual(len(list(reader[13:20])), 7)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[18:20])), 2)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[19:20])), 1)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[20:20])), 0)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[21:20])), 0)
        self.assertEqual(reader.position, 21)

        with self.assertRaises(StopIteration):
            next(reader)


class TestRemoteNotificationLog(NotificationLogTestCase):
    use_named_temporary_file = True

    def test_remote_notification_log(self):
        num_notifications = 42

        section_size = 5

        # Build a notification log (fixture).
        self.append_notifications(num_notifications)

        # Start a simple server.
        from wsgiref.util import setup_testing_defaults
        from wsgiref.simple_server import make_server

        port = 8080
        base_url = 'http://127.0.0.1:{}/notifications/'.format(port)

        def simple_app(environ, start_response):
            """Simple WSGI application."""
            setup_testing_defaults(environ)

            # Identify log and section from request.
            path_info = environ['PATH_INFO']
            try:
                section_id = path_info.strip('/').split('/')[-1]
            except ValueError:
                # Start response.
                status = '404 Not Found'
                headers = [('Content-type', 'text/plain; charset=utf-8')]
                start_response(status, headers)
                return []

            # Select the notification log.
            notification_log = self.create_notification_log(section_size)

            # Get serialized section.
            view = NotificationLogView(notification_log)
            section, is_archived = view.present_section(section_id)
            # Todo: Maybe redirect if the section ID is a mismatch, so
            # the URL is good for cacheing.

            # Start response.
            status = '200 OK'
            headers = [('Content-type', 'text/plain; charset=utf-8')]
            start_response(status, headers)

            # Return a list of lines.
            return [(line + '\n').encode('utf8') for line in section.split('\n')]

        httpd = make_server('', port, simple_app)
        print("Serving on port {}...".format(port))
        thread = Thread(target=httpd.serve_forever)
        thread.setDaemon(True)
        thread.start()
        try:
            # Use reader with client to read all items in remote feed after item 5.
            notification_log = RemoteNotificationLog(base_url)

            # Get all the items.
            notification_log_reader = NotificationLogReader(notification_log=notification_log)
            items_from_start = notification_log_reader.read_list()

            # Check we got all the items.
            self.assertEqual(len(items_from_start), num_notifications)
            self.assertEqual(items_from_start[0]['id'], 1)
            self.assertEqual(items_from_start[0]['data'], 'item1')
            self.assertEqual(items_from_start[0]['topic'], 'eventsourcing.domain.model.events#DomainEvent')
            expected_section_count = ceil(num_notifications / float(section_size))
            self.assertEqual(notification_log_reader.section_count, expected_section_count)

            # Get all the items from item 5.
            items_from_5 = list(notification_log_reader[section_size - 1:])

            # Check we got everything after item 5.
            self.assertEqual(len(items_from_5), num_notifications - section_size + 1)
            self.assertEqual(items_from_5[0]['id'], section_size)
            self.assertEqual(items_from_5[0]['data'], 'item{}'.format(section_size))
            self.assertEqual(items_from_5[0]['topic'], 'eventsourcing.domain.model.events#DomainEvent')
            expected_section_count = ceil(num_notifications / float(section_size))
            self.assertEqual(notification_log_reader.section_count, expected_section_count)

            # Check ValueError is raised for deserialization errors.
            with self.assertRaises(ValueError):
                notification_log.deserialize_section('invalid json')

        finally:
            httpd.shutdown()
            thread.join()
            httpd.server_close()


class TestNotificationLogWithDjango(DjangoTestCase, TestNotificationLog):
    pass


class TestBigArrayNotificationLogWithDjango(DjangoTestCase, TestBigArrayNotificationLog):
    pass


class TestBigArrayNotificationLogWithCassandra(WithCassandraRecordManagers, TestBigArrayNotificationLog):
    pass



from unittest import TestCase
from uuid import uuid4

from eventsourcing.application.sqlalchemy import CommandProcess, ProcessApplication, ProcessApplicationWithSnapshotting
from eventsourcing.application.process import RepositoryWrapper
from eventsourcing.domain.model.aggregate import BaseAggregateRoot, AggregateRoot
from eventsourcing.domain.model.command import Command
from eventsourcing.domain.model.events import assert_event_handlers_empty, subscribe, unsubscribe, \
    EventHandlersNotEmptyError, clear_event_handlers
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.exceptions import CausalDependencyFailed, PromptFailed
from eventsourcing.utils.topic import resolve_topic
from eventsourcing.utils.transcoding import json_loads


class TestProcess(TestCase):

    def test_process_with_example_policy(self):
        # Construct example process.
        with ProcessApplication(
            name='test',
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
        ) as  process:

            # Make the process follow itself.
            process.follow('test', process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            # Check the aggregate has been automatically "moved on".
            self.assertTrue(process.repository[aggregate.id].is_moved_on)

            # Check the __contains__ method of the repo wrapper.
            self.assertIn(aggregate.id, RepositoryWrapper(process.repository))
            self.assertNotIn(uuid4(), RepositoryWrapper(process.repository))

            # Check the repository wrapper tracks causal dependencies.
            repository = RepositoryWrapper(process.repository)
            aggregate = repository[aggregate.id]
            causal_dependencies = repository.causal_dependencies
            self.assertEqual(len(causal_dependencies), 1)
            self.assertEqual((aggregate.id, 1), causal_dependencies[0])

            # Check events from more than one aggregate are stored.
            self.assertIn(aggregate.second_id, process.repository)

    def test_process_application_with_snapshotting(self):
        # Construct example process.
        with ProcessApplicationWithSnapshotting(
            name='test',
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
            snapshot_period=2,
        ) as process:

            # Make the process follow itself.
            process.follow('test', process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()

            # Check there isn't a snapshot.
            self.assertIsNone(process.snapshot_strategy.get_snapshot(aggregate.id))

            # Should "move on" by the process following itself.
            aggregate.__save__()
            aggregate = process.repository[aggregate.id]
            self.assertEqual(1, aggregate.__version__)

            # Check there is a snapshot.
            snapshot = process.snapshot_strategy.get_snapshot(aggregate.id)
            self.assertIsInstance(snapshot, Snapshot)
            self.assertEqual(snapshot.originator_version, 1)

            snapshot_v0 = process.snapshot_strategy.get_snapshot(aggregate.id, lt=snapshot.originator_version)
            self.assertIsNone(snapshot_v0, Snapshot)

    def test_causal_dependencies(self):
        # Try to process an event that has unresolved causal dependencies.
        pipeline_id1 = 0
        pipeline_id2 = 1

        # Create two events, one has causal dependency on the other.
        core1 = ProcessApplication(
            name='core',
            # persist_event_type=ExampleAggregate.Created,
            persist_event_type=BaseAggregateRoot.Event,
            setup_table=True,
            pipeline_id=pipeline_id1,
        )

        core2 = ProcessApplication(
            name='core',
            pipeline_id=pipeline_id2,
            policy=example_policy,
            session=core1.session
        )

        # First event in pipeline 1.
        aggregate = ExampleAggregate.__create__()
        aggregate.__save__()

        # Second event in pipeline 2.
        # - it's important this is done in a policy so the causal dependencies are identified
        core2.follow('core', core1.notification_log)
        core2.run()

        # Check the aggregate exists.
        self.assertTrue(aggregate.id in core1.repository)

        # Check the aggregate has been "moved on".
        aggregate = core1.repository[aggregate.id]
        self.assertTrue(aggregate.is_moved_on)
        self.assertTrue(aggregate.second_id)
        self.assertIn(aggregate.second_id, core1.repository)

        # Check the events have different pipeline IDs.
        aggregate_records = core1.event_store.record_manager.get_records(aggregate.id)
        second_entity_records = core1.event_store.record_manager.get_records(aggregate.second_id)

        self.assertEqual(2, len(aggregate_records))
        self.assertEqual(1, len(second_entity_records))

        self.assertEqual(pipeline_id1, aggregate_records[0].pipeline_id)
        self.assertEqual(pipeline_id2, aggregate_records[1].pipeline_id)
        self.assertEqual(pipeline_id2, second_entity_records[0].pipeline_id)

        # Check the causal dependencies have been constructed.
        # - the first 'Created' event doesn't have an causal dependencies
        self.assertEqual(None, aggregate_records[0].causal_dependencies)

        # - the second 'Created' event depends on the Created event in another pipeline.
        expect = [{
            'notification_id': 1,
            'pipeline_id': pipeline_id1
        }]
        actual = json_loads(second_entity_records[0].causal_dependencies)

        self.assertEqual(expect, actual)

        # - the 'AttributeChanged' event depends on the second Created,
        # which is in the same pipeline, so expect no causal dependencies.
        self.assertEqual(None, aggregate_records[1].causal_dependencies)

        # Setup downstream process.
        downstream1 = ProcessApplication(
            'downstream',
            pipeline_id=pipeline_id1,
            session=core1.session,
            policy=event_logging_policy,
        )
        downstream1.follow('core', core1.notification_log)
        downstream2 = ProcessApplication(
            'downstream',
            pipeline_id=pipeline_id2,
            session=core1.session,
            policy=event_logging_policy,
        )
        downstream2.follow('core', core2.notification_log)

        # Try to process pipeline 2, should fail due to causal dependency.
        with self.assertRaises(CausalDependencyFailed):
            downstream2.run()

        self.assertEqual(0, len(downstream1.event_store.record_manager.get_notifications()))
        self.assertEqual(0, len(downstream2.event_store.record_manager.get_notifications()))

        # Try to process pipeline 1, should work.
        downstream1.run()

        self.assertEqual(1, len(downstream1.event_store.record_manager.get_notifications()))
        self.assertEqual(0, len(downstream2.event_store.record_manager.get_notifications()))

        # Try again to process pipeline 2, should work this time.
        downstream2.run()

        self.assertEqual(1, len(downstream1.event_store.record_manager.get_notifications()))
        self.assertEqual(2, len(downstream2.event_store.record_manager.get_notifications()))

        core1.close()
        core2.close()
        downstream1.close()
        downstream2.close()

    def test_handle_prompt_failed(self):
        process = ProcessApplication(
            name='test',
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
        )

        def raise_exception(_):
            raise Exception()

        def raise_prompt_failed(_):
            raise PromptFailed()

        subscribe(raise_exception)
        try:
            with self.assertRaises(PromptFailed):
                process.publish_prompt()
        finally:
            unsubscribe(raise_exception)

        subscribe(raise_prompt_failed)
        try:
            with self.assertRaises(PromptFailed):
                process.publish_prompt()
        finally:
            unsubscribe(raise_prompt_failed)

        try:
            process.publish_prompt()
        finally:
            process.close()

    def tearDown(self):
        try:
            assert_event_handlers_empty()
        except EventHandlersNotEmptyError:
            clear_event_handlers()
            raise


class TestCommands(TestCase):

    def test_command_aggregate(self):
        # Create a command.
        cmd = Command.__create__()
        assert isinstance(cmd, Command)

        # Check is_done gets changed by done()
        self.assertFalse(cmd.is_done)
        cmd.done()
        self.assertTrue(cmd.is_done)

        # Check the resulting domain events.
        pending_events = cmd.__batch_pending_events__()
        self.assertIsInstance(pending_events[0], Command.Created)
        self.assertIsInstance(pending_events[1], Command.Done)

    def test_command_process(self):
        commands = CommandProcess(
            setup_table=True
        )
        core = ProcessApplication(
            'core',
            policy=example_policy,
            session=commands.session
        )

        self.assertFalse(list(commands.event_store.all_domain_events()))

        cmd = CreateExample.__create__()
        # cmd = Command.__create__(cmd_method='create_example', cmd_args={})
        cmd.__save__()

        domain_events = list(commands.event_store.all_domain_events())
        # self.assertTrue(domain_events)
        self.assertEqual(len(domain_events), 1)

        self.assertFalse(list(core.event_store.all_domain_events()))

        core.follow('commands', commands.notification_log)
        core.run()

        self.assertTrue(list(core.event_store.all_domain_events()))

        commands.close()
        core.close()

    def tearDown(self):
        assert_event_handlers_empty()


# Example aggregate (used in the test).

class ExampleAggregate(BaseAggregateRoot):
    def __init__(self, **kwargs):
        super(ExampleAggregate, self).__init__(**kwargs)
        self.is_moved_on = False
        self.second_id = None

    class Event(BaseAggregateRoot.Event):
        pass

    class Created(Event, BaseAggregateRoot.Created):
        pass

    def move_on(self, second_id=None):
        self.__trigger_event__(ExampleAggregate.MovedOn, second_id=second_id)

    class MovedOn(Event):
        @property
        def second_id(self):
            return self.__dict__['second_id']

        def mutate(self, aggregate):
            assert isinstance(aggregate, ExampleAggregate)
            aggregate.is_moved_on = True
            aggregate.second_id = self.second_id


def example_policy(process, repository, event):
    # Whenever an aggregate is created, then "move it on".
    if isinstance(event, ExampleAggregate.Created):
        # Get aggregate and move it on.
        aggregate = repository[event.originator_id]

        assert isinstance(aggregate, ExampleAggregate)

        # Also create a second entity, allows test to check that
        # events from more than one entity are stored.
        second_id = uuid4()
        other_entity = AggregateRoot.__create__(originator_id=second_id)
        aggregate.move_on(second_id=second_id)
        return other_entity

    elif isinstance(event, Command.Created):
        command_class = resolve_topic(event.originator_topic)
        if command_class is CreateExample:
            return ExampleAggregate.__create__()


class LogMessage(BaseAggregateRoot):

    def __init__(self, message='', **kwargs):
        super(LogMessage, self).__init__(**kwargs)
        self.message = message

    class Event(BaseAggregateRoot.Event):
        pass

    class Created(Event, BaseAggregateRoot.Created):
        pass


def event_logging_policy(process, repository, event):
    return LogMessage.__create__(uuid4(), message=str(event))


class CreateExample(Command):
    pass



from eventsourcing.tests.datastore_tests.test_cassandra import CassandraDatastoreTestCase
from eventsourcing.tests.sequenced_item_tests.base import IntegerSequencedRecordTestCase, \
    SimpleSequencedItemteratorTestCase, ThreadedSequencedItemIteratorTestCase, TimestampSequencedItemTestCase, \
    WithRecordManagers


class TestCassandraRecordManagerWithIntegerSequences(CassandraDatastoreTestCase,
                                                     IntegerSequencedRecordTestCase):
    pass


class TestCassandraRecordManagerWithTimestampSequences(CassandraDatastoreTestCase,
                                                       TimestampSequencedItemTestCase):
    pass


class WithCassandraRecordManagers(CassandraDatastoreTestCase, WithRecordManagers):
    pass


class TestSimpleSequencedItemIteratorWithCassandra(WithCassandraRecordManagers,
                                                   SimpleSequencedItemteratorTestCase):
    pass


class TestThreadedSequencedItemIteratorWithCassandra(WithCassandraRecordManagers,
                                                     ThreadedSequencedItemIteratorTestCase):
    pass



import os
from setuptools import find_packages, setup

try:
    from functools import singledispatch

    singledispatch_requires = []
except ImportError:
    singledispatch_requires = ['singledispatch==3.4.0.3']

from eventsourcing import __version__

# Read the docs doesn't need to build the Cassandra driver (and can't).
if 'READTHEDOCS' in os.environ:
    os.environ['CASS_DRIVER_NO_CYTHON'] = '1'

install_requires = singledispatch_requires + [
    'python-dateutil<=2.6.99999',
    'six<=1.11.99999',
    'pycryptodome<=3.4.99999',
    'requests<=2.18.99999',
]

sqlalchemy_requires = [
    'sqlalchemy<=1.2.99999,>=0.9',
    'sqlalchemy-utils<=0.32.99999',
]

cassandra_requires = [
    'cassandra-driver<=3.12.99999'
]

django_requires = [  # Note, Django 2 doesn't support Python 2.7
    'django>=1.11,<=2.0.99999' if str != bytes else 'django>=1.11,<=1.11.99999',
]

testing_requires = [
    'mock<=2.0.99999',
    'flask<=0.12.99999',
    'flask_sqlalchemy<=2.3.99',
    'uwsgi<=2.0.99999',
    'redis<=2.10.99999',
    'celery<=4.1.99999',
    'pymysql<=0.8.99999',
    'thespian<=3.9.99999',
] + cassandra_requires + sqlalchemy_requires + django_requires

docs_requires = ['Sphinx', 'sphinx_rtd_theme', 'sphinx-autobuild'] + testing_requires


long_description = """
A library for event sourcing in Python.

`Package documentation is now available <http://eventsourcing.readthedocs.io/>`_.

`Please raise issues on GitHub <https://github.com/johnbywater/eventsourcing/issues>`_.
"""

packages = find_packages(
    exclude=[
        "docs",
        # "eventsourcing.contrib*",
        # "eventsourcing.tests*"
    ]
)

setup(
    name='eventsourcing',
    version=__version__,
    description='Event sourcing in Python',
    author='John Bywater',
    author_email='john.bywater@appropriatesoftware.net',
    url='https://github.com/johnbywater/eventsourcing',
    packages=packages,
    install_requires=install_requires,
    extras_require={
        'cassandra': cassandra_requires,
        'sqlalchemy': sqlalchemy_requires,
        'django': django_requires,
        'test': testing_requires,
        'tests': testing_requires,
        'testing': testing_requires,
        'dev': testing_requires,
        'develop': testing_requires,
        'development': testing_requires,
        'docs': docs_requires,
    },
    zip_safe=False,
    long_description=long_description,
    keywords=['event sourcing', 'event store', 'domain driven design', 'ddd', 'cqrs', 'cqs'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from __future__ import absolute_import, division, print_function, unicode_literals

import json
from abc import ABCMeta, abstractmethod

import requests
import six

from eventsourcing.domain.model.array import BigArray
from eventsourcing.infrastructure.base import ACIDRecordManager
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder, json_dumps

DEFAULT_SECTION_SIZE = 20


class Section(object):
    """
    Section of a notification log.

    Contains items, and has an ID.

    May also have either IDs of previous and next sections of the notification log.
    """

    def __init__(self, section_id, items, previous_id=None, next_id=None):
        self.section_id = section_id
        self.items = items
        self.previous_id = previous_id
        self.next_id = next_id


class AbstractNotificationLog(six.with_metaclass(ABCMeta)):
    """
    Presents a sequence of sections from a sequence of notifications.
    """

    @abstractmethod
    def __getitem__(self, section_id):
        """
        Get section of notification log.

        :rtype: Section
        """


class LocalNotificationLog(AbstractNotificationLog):
    """
    Presents a sequence of sections from a sequence of notifications.
    """

    def __init__(self, section_size=None):
        self.section_size = section_size or DEFAULT_SECTION_SIZE
        # self.last_start = None

    def __getitem__(self, section_id):
        # Get section of notification log.
        next_position = None
        if section_id == 'current':

            # Get start of the last section.
            next_position = self.get_next_position()
            start = next_position // self.section_size * self.section_size
        else:

            try:
                first_item_number, last_item_number = section_id.split(',')
            except ValueError as e:
                raise ValueError("Couldn't split '{}': {}".format(section_id, e))

            # Convert from 1-based to 0-based.
            start = int(first_item_number) - 1

            # Go back to nearest section start.
            start = start // self.section_size * self.section_size

        # Get stop index (section doesn't include this one).
        stop = start + self.section_size

        # Get the items.
        items = list(self.get_items(start, stop, next_position))

        # Get previous and next section IDs.
        if start:
            first_item_number = start + 1 - self.section_size
            last_item_number = first_item_number - 1 + self.section_size
            previous_id = self.format_section_id(first_item_number, last_item_number)
        else:
            previous_id = None

        if len(items) == self.section_size:
            first_item_number = start + 1 + self.section_size
            last_item_number = first_item_number - 1 + self.section_size
            next_id = self.format_section_id(first_item_number, last_item_number)
        else:
            next_id = None

        # Return a section of the notification log.
        section_id = self.format_section_id(start + 1, start + self.section_size)
        return Section(
            section_id=section_id,
            items=items,
            previous_id=previous_id,
            next_id=next_id,
        )

    @abstractmethod
    def get_next_position(self):
        """
        Returns items for section.

        :rtype: int
        """

    @abstractmethod
    def get_items(self, start, stop, next_position=None):
        """
        Returns items for section.

        :rtype: list
        """

    @staticmethod
    def format_section_id(first_item_number, last_item_number):
        return '{},{}'.format(first_item_number, last_item_number)


class RecordManagerNotificationLog(LocalNotificationLog):
    def __init__(self, record_manager, section_size):
        super(RecordManagerNotificationLog, self).__init__(section_size)
        assert isinstance(record_manager, ACIDRecordManager), record_manager
        assert record_manager.contiguous_record_ids
        self.record_manager = record_manager

    def get_items(self, start, stop, next_position=None):
        notifications = []
        for record in self.record_manager.get_notifications(start, stop):
            notification = {'id': record.id}
            for field_name in self.record_manager.field_names:
                notification[field_name] = getattr(record, field_name)
            if hasattr(record, 'causal_dependencies'):
                notification['causal_dependencies'] = record.causal_dependencies
            notifications.append(notification)
        return notifications

    def get_next_position(self):
        return self.get_end_position() or 1

    def get_end_position(self):
        return self.record_manager.get_max_record_id() or 0


class BigArrayNotificationLog(LocalNotificationLog):
    def __init__(self, big_array, section_size):
        super(BigArrayNotificationLog, self).__init__(section_size)
        assert isinstance(big_array, BigArray)
        if big_array.repo.array_size % section_size:
            raise ValueError("Section size {} doesn't divide array size {}".format(
                section_size, big_array.repo.array_size
            ))
        self.big_array = big_array

    def get_items(self, start, stop, next_position=None):
        next_position = self.get_next_position() if next_position is None else next_position
        stop = min(stop, next_position)
        return self.big_array[start:stop]

    def get_next_position(self):
        return self.big_array.get_next_position()


class RemoteNotificationLog(AbstractNotificationLog):
    def __init__(self, base_url, json_decoder_class=None):
        self.base_url = base_url
        self.json_decoder_class = json_decoder_class

    def __getitem__(self, section_id):
        section_json = self.get_json(section_id)
        return self.deserialize_section(section_json)

    def deserialize_section(self, section_json):
        try:
            decoder_class = self.json_decoder_class or ObjectJSONDecoder
            section = Section(**json.loads(section_json, cls=decoder_class))
        except ValueError as e:
            raise ValueError("Couldn't deserialize notification log section: "
                             "{}: {}".format(e, section_json))
        return section

    def get_json(self, section_id):
        notification_log_url = self.make_notification_log_url(section_id)
        return self.get_resource(notification_log_url)

    def get_resource(self, doc_url):
        representation = requests.get(doc_url).content
        if isinstance(representation, type(b'')):
            representation = representation.decode('utf8')
        return representation

    def make_notification_log_url(self, section_id):
        return '{}/{}/'.format(self.base_url.strip('/'), section_id)


class NotificationLogReader(six.with_metaclass(ABCMeta)):
    def __init__(self, notification_log):
        assert isinstance(notification_log, AbstractNotificationLog)
        self.notification_log = notification_log
        self.section_count = 0
        self.position = 0

    def __getitem__(self, item=None):
        if isinstance(item, slice):
            assert item.start >= 0, item.start
            self.seek(item.start)
            return self.read_items(item.stop)
        else:
            assert isinstance(item, int), type(item)
            assert item >= 0, item
            self.seek(item)
            return self.read_list(advance_by=1)[0]

    def __iter__(self):
        return self.read_items()

    def __next__(self):
        try:
            return self.read_list(advance_by=1)[0]
        except IndexError:
            raise StopIteration

    def next(self):
        """
        Python 2.7 version of the iterator protocol.
        """
        return self.__next__()

    def seek(self, position):
        if position < 0:
            raise ValueError("Position less than zero: {}".format(position))
        self.position = position

    def read(self, advance_by=None):
        return self.read_items(advance_by=advance_by)

    def read_items(self, stop_index=None, advance_by=None):
        self.section_count = 0

        start_item_num = self.position + 1

        # Validate the position.
        if self.position < 0:
            raise ValueError("Position less than zero: {}".format(self.position))

        # Get current section.
        section = self.notification_log['current']

        # Follow previous links.
        while section.previous_id:

            # Break if we can go forward from here.
            if start_item_num is not None:
                if int(section.section_id.split(',')[0]) <= start_item_num:
                    break

            # Get the previous document.
            section_id = section.previous_id
            section = self.notification_log[section_id]

        # Yield items in first section, optionally after last item number.
        self.section_count += 1
        items = section.items
        if start_item_num is not None:
            section_start_num = int(section.section_id.split(',')[0])
            from_index = start_item_num - section_start_num
            items = items[from_index:]

        if advance_by is None:
            advance_by = -1

        # Yield all items in all subsequent sections.
        while True:

            items_iter = iter(items)
            while True:
                if stop_index is not None and self.position >= stop_index:
                    return
                if advance_by == 0:
                    return
                try:
                    item = next(items_iter)
                except StopIteration:
                    break
                self.position += 1
                advance_by -= 1
                yield item

            if section.next_id:
                # Follow link to get next section.
                section = self.notification_log[section.next_id]
                items = section.items
                self.section_count += 1
            else:
                break

    def read_list(self, advance_by=None):
        return list(self.read_items(advance_by=advance_by))


class NotificationLogView(object):
    def __init__(self, notification_log, json_encoder_class=None):
        assert isinstance(notification_log, LocalNotificationLog), type(notification_log)
        self.notification_log = notification_log
        self.json_encoder_class = json_encoder_class or ObjectJSONEncoder

    def present_section(self, section_id):
        section = self.notification_log[section_id]
        is_archived = bool(section.next_id)
        section_json = json_dumps(section.__dict__, self.json_encoder_class)
        return section_json, is_archived



from abc import ABCMeta, abstractmethod

import six


class DatastoreSettings(object):
    """Base class for settings for database connection used by a stored event repository."""


class Datastore(six.with_metaclass(ABCMeta)):
    def __init__(self, settings):
        # assert isinstance(settings, DatastoreSettings), settings
        self.settings = settings

    @abstractmethod
    def setup_connection(self):
        """Sets up a connection to a datastore."""

    @abstractmethod
    def close_connection(self):
        """Drops connection to a datastore."""

    @abstractmethod
    def setup_tables(self):
        """Sets up tables used to store events."""

    @abstractmethod
    def drop_tables(self):
        """Drops tables used to store events."""

    @abstractmethod
    def truncate_tables(self):
        """Truncates tables used to store events."""


class DatastoreError(Exception):
    pass


class DatastoreConnectionError(DatastoreError):
    pass


class DatastoreTableError(DatastoreError):
    pass
